from PyQt5 import QtWidgets, QtCore, QtGui

from skippy.gui import utils

from skippy.utils import translator

import skippy.config

from typing import Optional, Dict
import os


class FilesDialog(QtWidgets.QDialog):
    def __init__(
        self, files: Dict[str, str], parent: Optional[QtWidgets.QWidget] = None
    ):
        super(FilesDialog, self).__init__(parent)
        self._layout = QtWidgets.QVBoxLayout(self)

        self.label = QtWidgets.QLabel(
            translator.Translator().translate("DIALOG.FILES_LIST_LABEL"), self
        )
        self.label.setFont(QtGui.QFont("Arial", 10))

        self.scrollArea = QtWidgets.QScrollArea()
        self.scrollArea.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        self.scrollArea.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.scrollArea.setWidgetResizable(True)

        self.fileWidget = FilesWidget(files, self)
        self.fileWidget.filesListUpdated.connect(self.update)
        self.scrollArea.setWidget(self.fileWidget)

        self._layout.addWidget(self.label, alignment=QtCore.Qt.AlignCenter)
        self._layout.addWidget(self.scrollArea, alignment=QtCore.Qt.AlignCenter)

        self.setLayout(self._layout)

        self.move(300, 300)
        self.resize(300, 200)
        self.setWindowTitle(f"Skippy - {skippy.config.version}")

        self.show()


class FilesWidget(QtWidgets.QWidget):
    filesListUpdated = QtCore.pyqtSignal()

    def __init__(
        self, files: Dict[str, str], parent: Optional[QtWidgets.QWidget] = None
    ):
        super(FilesWidget, self).__init__(parent)
        self.files = files
        self._layout = QtWidgets.QVBoxLayout(self)

        self.fileWidgets = {}
        for title in self.files:
            fileWidget = FileLineEdit(title, self)
            fileWidget.fileNameChanged.connect(self.updateFileName)
            fileWidget.fileRemoved.connect(self.removeFile)
            self._layout.addWidget(fileWidget)
            self.fileWidgets[title] = fileWidget

        self.setLayout(self._layout)

    def updateFileName(self, old: str, new: str):
        self.files[new] = self.files.pop(old)

    def removeFile(self, file: str):
        del self.files[file]
        self.fileWidgets[file].setParent(None)

        self.resize(self.sizeHint())


class FileLineEdit(QtWidgets.QLineEdit):
    fileNameChanged = QtCore.pyqtSignal(str, str)
    fileRemoved = QtCore.pyqtSignal(str)

    def __init__(self, title: str, parent: Optional[QtWidgets.QWidget] = None):
        super(FileLineEdit, self).__init__(title, parent)
        theme = utils.getMainWindow().settings.theme

        self._prevText = title

        self.button = QtWidgets.QToolButton(self)
        self.button.setIcon(
            QtGui.QIcon(
                os.path.join(skippy.config.RESOURCES_FOLDER, theme, "close.png")
            )
        )
        self.button.setStyleSheet("padding: 0px;")
        self.button.setCursor(QtCore.Qt.ArrowCursor)
        self.button.clicked.connect(lambda: self.fileRemoved.emit(self.text()))

        frameWidth = self.style().pixelMetric(QtWidgets.QStyle.PM_DefaultFrameWidth)
        buttonSize = self.button.sizeHint()

        self.textEdited.connect(self.textEditedEvent)

        self.setStyleSheet(
            "QLineEdit {padding-right: "
            + str(buttonSize.width() + frameWidth + 1)
            + "px; }"
        )
        self.setFixedSize(200, 40)

    def textEditedEvent(self, new_text: str):
        self.fileNameChanged.emit(self._prevText, new_text)
        self._prevText = new_text

    def resizeEvent(self, event: QtGui.QResizeEvent):
        buttonSize = self.button.sizeHint()
        frameWidth = self.style().pixelMetric(QtWidgets.QStyle.PM_DefaultFrameWidth)
        self.button.move(
            self.rect().right() - frameWidth - buttonSize.width(),
            (self.rect().bottom() - buttonSize.height() + 1) / 2,
        )
        super(FileLineEdit, self).resizeEvent(event)



from PyQt5 import QtWidgets, QtCore, QtGui

from skippy.api import PageData

from skippy.gui import sitebox, workers, thread

from skippy.utils import translator
import skippy.config

from typing import Optional
import os


class UploadDialog(QtWidgets.QDialog):
    def __init__(self, pdata: PageData, parent: Optional[QtWidgets.QWidget] = None):
        super(UploadDialog, self).__init__(parent)
        self.pdata = pdata

        self._layout = QtWidgets.QVBoxLayout(self)

        self.label = QtWidgets.QLabel(
            translator.Translator().translate("DIALOG.ENTER_PAGE_LABEL"), self
        )

        self.site_box = sitebox.SiteBox(
            self, self.pdata["link"][0] if self.pdata["link"] else ""
        )

        self.page_box = QtWidgets.QLineEdit()
        self.page_box.setPlaceholderText(
            translator.Translator().translate("DIALOG.PAGE_BOX_PLACEHOLDER")
        )
        self.page_box.setText(self.pdata["link"][1] if self.pdata["link"] else "")

        self.comment_box = QtWidgets.QLineEdit()
        self.comment_box.setText(
            translator.Translator().translate("DIALOG.COMMENT_BOX_TEXT")
        )
        self.comment_box.setPlaceholderText(
            translator.Translator().translate("DIALOG.COMMENT_BOX_PLACEHOLDER")
        )

        self.button = QtWidgets.QPushButton(
            translator.Translator().translate("DIALOG.OK_BUTTON"), self
        )
        self.button.clicked.connect(self.upload)

        self._layout.addWidget(self.label, alignment=QtCore.Qt.AlignCenter)
        self._layout.addWidget(self.site_box)
        self._layout.addWidget(self.page_box)
        self._layout.addWidget(self.comment_box)
        self._layout.addWidget(self.button, alignment=QtCore.Qt.AlignRight)

        self.setLayout(self._layout)

        self.setWindowTitle(f"Skippy - {skippy.config.version}")
        self.setWindowIcon(
            QtGui.QIcon(os.path.join(skippy.config.RESOURCES_FOLDER, "skippy.ico"))
        )
        self.move(300, 300)
        self.resize(200, 100)

        self.show()

    def upload(self):
        site = self.site_box.currentText()
        page = self.page_box.text()
        if site and page:
            comment = self.comment_box.text()

            self.pdata["link"] = (site, page)

            self._thread = thread.Thread(workers.UploadWorker(self.pdata, comment))
            self._thread.start()

        self.close()



from PyQt5 import QtWidgets

from skippy.utils import translator, filehandlers

from typing import Optional


class LoginStatus(QtWidgets.QLabel):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super(LoginStatus, self).__init__(parent)
        self.setFixedWidth(self.width() + 100)
        self.updateStatus(filehandlers.ProfileHandler().load()[0])

    def updateStatus(self, username: str):
        if username:
            self.setText(
                translator.Translator()
                .translate("MAIN.SIGNED_IN_AS_LABEL")
                .format(username)
            )
        else:
            self.setText("")



from PyQt5 import QtWidgets, QtGui

from skippy.gui.utils import getMainWindow

from skippy.utils import translator

import skippy.config

from typing import Optional
import re
import os


class FinderDialog(QtWidgets.QDialog):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super(FinderDialog, self).__init__(parent)
        self.textEditor = getMainWindow().tab.currentWidget().editor

        self.lastStart = 0

        findButton = QtWidgets.QPushButton(
            translator.Translator().translate("DIALOG.FINDER_FIND_BUTTON"), self
        )
        findButton.clicked.connect(self.find)

        replaceButton = QtWidgets.QPushButton(
            translator.Translator().translate("DIALOG.FINDER_REPLACE_BUTTON"), self
        )
        replaceButton.clicked.connect(self.replace)

        allButton = QtWidgets.QPushButton(
            translator.Translator().translate("DIALOG.FINDER_REPLACE_ALL_BUTTON"), self
        )
        allButton.clicked.connect(self.replaceAll)

        self.normalRadio = QtWidgets.QRadioButton(
            translator.Translator().translate("DIALOG.FINDER_NORMAL_MODE"), self
        )

        regexRadio = QtWidgets.QRadioButton(
            translator.Translator().translate("DIALOG.FINDER_REGEX_MODE"), self
        )

        self.findField = QtWidgets.QTextEdit(self)
        self.findField.resize(250, 50)

        self.replaceField = QtWidgets.QTextEdit(self)
        self.replaceField.resize(250, 50)

        layout = QtWidgets.QGridLayout()

        layout.addWidget(self.findField, 1, 0, 1, 4)
        layout.addWidget(self.normalRadio, 2, 2)
        layout.addWidget(regexRadio, 2, 3)
        layout.addWidget(findButton, 2, 0, 1, 2)

        layout.addWidget(self.replaceField, 3, 0, 1, 4)
        layout.addWidget(replaceButton, 4, 0, 1, 2)
        layout.addWidget(allButton, 4, 2, 1, 2)

        self.setGeometry(300, 300, 360, 250)
        self.setWindowTitle(f"Skippy - {skippy.config.version}")
        self.setWindowIcon(
            QtGui.QIcon(os.path.join(skippy.config.RESOURCES_FOLDER, "skippy.ico"))
        )
        self.setLayout(layout)

        self.normalRadio.setChecked(True)
        self.show()

    def find(self):
        text = self.textEditor.toPlainText()
        query = self.findField.toPlainText()

        if self.normalRadio.isChecked():
            self.lastStart = text.find(query, self.lastStart + 1)

            if self.lastStart >= 0:
                end = self.lastStart + len(query)
                self.moveCursor(self.lastStart, end)
            else:
                self.lastStart = 0
        else:
            pattern = re.compile(query)

            match = pattern.search(text, self.lastStart + 1)
            if match:
                self.lastStart = match.start()
                self.moveCursor(self.lastStart, match.end())
            else:
                self.lastStart = 0

    def replace(self):
        self.find()
        cursor = self.textEditor.textCursor()

        if cursor.hasSelection():
            cursor.insertText(self.replaceField.toPlainText())
            cursor.select(QtGui.QTextCursor.WordUnderCursor)
            self.textEditor.setTextCursor(cursor)

    def replaceAll(self):
        self.lastStart = 0
        self.replace()

        while self.lastStart:
            self.replace()

    def moveCursor(self, start, end):
        cursor = self.textEditor.textCursor()
        cursor.setPosition(start)
        cursor.movePosition(
            QtGui.QTextCursor.Right, QtGui.QTextCursor.KeepAnchor, end - start
        )

        self.textEditor.setTextCursor(cursor)



from skippy.gui import CriticalMessageBox

from skippy.utils.logger import log

import traceback
import sys


def excepthook(exc_type, value, tb):
    exc = "".join(traceback.format_exception(exc_type, value, tb))
    log.error(exc)
    CriticalMessageBox(f"{exc_type.__name__}: {str(value)}", exc).exec_()


def init():
    sys.excepthook = excepthook



from PyQt5 import QtWidgets, QtCore

from skippy.core.elements import AbstractElement

from skippy.gui.utils import getMainWindow

from skippy.utils.translator import Translator

import skippy.config

from typing import Optional


class ElementDialog(QtWidgets.QDialog):
    def __init__(
        self, element: AbstractElement, parent: Optional[QtWidgets.QWidget] = None
    ):
        super(ElementDialog, self).__init__(parent)
        self.element = element

        self._layout = QtWidgets.QVBoxLayout(self)

        self.element_data = QtWidgets.QLabel(
            Translator().translate(self.element.__alias__)
            + " - "
            + Translator().translate(self.element.__description__),
            self,
        )
        self.element_data.setWordWrap(True)

        self._layout.addWidget(self.element_data)
        self._layout.addWidget(QHLine())

        self.scroll = QtWidgets.QScrollArea()
        self.widget = QtWidgets.QWidget()
        self.vbox = QtWidgets.QVBoxLayout()
        self.fields = {}
        for field in self.element.required_fields:
            label = QtWidgets.QLabel(
                Translator().translate(field.name) if field.name else field.tag, self
            )

            line = QtWidgets.QPlainTextEdit(self)
            line.setFixedHeight(25)
            line.setPlaceholderText(Translator().translate(field.description))
            line.textChanged.connect(self.updatePreview)
            self.fields[field.tag] = line

            self.vbox.addWidget(label)
            self.vbox.addWidget(line)

        self.widget.setLayout(self.vbox)

        self.scroll.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        self.scroll.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.scroll.setWidgetResizable(True)
        self.scroll.setWidget(self.widget)

        self._layout.addWidget(self.scroll)

        self.previewer = QtWidgets.QPlainTextEdit(self)
        self.previewer.setReadOnly(True)

        self._layout.addWidget(self.previewer)

        self.button = QtWidgets.QPushButton(
            Translator().translate("DIALOG.GENERATE_BUTTON"), self
        )
        self.button.clicked.connect(self.process)
        self._layout.addWidget(self.button, alignment=QtCore.Qt.AlignRight)

        self.setLayout(self._layout)

        self.setWindowTitle(
            f"{Translator().translate(self.element.__alias__)} | skippy - {skippy.config.version}"
        )
        self.move(300, 300)
        self.resize(500, 400)

        self.updatePreview()

        self.show()

    def updatePreview(self):
        self.previewer.setPlainText(self.preview())

    def process(self):
        getMainWindow().tab.currentWidget().editor.insertPlainText(self.preview())
        self.close()

    def preview(self):
        return self.element.preview(
            {field: self.fields[field].toPlainText() for field in self.fields}
        )


class QHLine(QtWidgets.QFrame):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super(QHLine, self).__init__(parent)
        self.setFrameShape(QtWidgets.QFrame.HLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)



"""Plugin system core module
"""
from skippy.core.settings import AbstractSetting

from skippy.utils.logger import log

import skippy.config

from typing import Iterator, Optional, Callable, Union, Dict, List, Any
from abc import ABCMeta, abstractmethod
import importlib
import toml
import sys
import os

sys.path.append(skippy.config.PLUGINS_FOLDER)


def inject(
    module: str, attr: str, value: Optional[Any] = None
) -> Union[Any, Callable[[Any], Any]]:
    """Inject value into selected module attribute.

    Note: This function is not safe. It can be used only in plugins for injecting variable into skippy source code.

    Args:
        module (str): Module name
        attr (str): Attribute name
        value (Optional[Any], optional): Any value or None (if None func return decorator wrapper)

    Returns:
        Union[Any, Callable[Any]]: If value is not None return value, else return decorator wrapper
    """

    def inject(value: Any) -> Any:
        """Inject value into selected module attribute.

        Args:
            value (Any): Any value

        Returns:
            Any: Input value
        """
        loc = {}
        exec(f"import {module}; module = {module}", globals(), loc)
        setattr(loc["module"], attr, value)

        return value

    return inject(value) if value else inject


class AbstractPlugin(metaclass=ABCMeta):

    """Base plugin class"""

    __alias__: str

    __description__: str
    __author__: str
    __version__: str

    def __init__(self):
        self._settingsPath: str = os.path.join(
            skippy.config.PLUGINS_SETTINGS_FOLDER, f"{self.__alias__}.toml"
        )
        self._settings: Dict[str, AbstractSetting] = {}

    def addSetting(self, name: str, setting: AbstractSetting):
        self._settings[name] = setting

    def load_settings(self):
        if self._settings and os.path.exists(self._settingsPath):
            with open(self._settingsPath) as f:
                settings = toml.load(f)
            for setting in settings:
                if setting in self._settings:
                    self._settings[setting].from_toml(settings[setting])

    def save_settings(self):
        if self._settings:
            with open(self._settingsPath, "w") as f:
                toml.dump(
                    {
                        setting: self._settings[setting].to_toml()
                        for setting in self._settings
                    },
                    f,
                )

    @abstractmethod
    def start(self):
        """Abstract start method"""
        pass

    @abstractmethod
    def stop(self):
        """Abstract stop method"""
        pass


class PluginLoader:

    """Plugin loader class"""

    def __init__(self):
        """Init Plugin loader"""
        self._plugins: List[AbstractPlugin] = []

    @staticmethod
    def plugins() -> list:
        """Get list of all plugins

        Returns:
            list: List of plugins
        """
        return [
            os.path.splitext(file)[0]
            for file in os.listdir(skippy.config.PLUGINS_FOLDER)
            if os.path.isfile(os.path.join(skippy.config.PLUGINS_FOLDER, file))
            if file.endswith("_plugin.py")
        ]

    @classmethod
    def plugins_data(cls) -> Iterator[Dict[str, str]]:
        """Get data of all plugins

        Yields:
            Dict[str, str]: Dict of plugin data
        """
        for file in cls.plugins():
            plugin = cls.import_plugin(file)
            yield {
                "__alias__": plugin.__alias__,
                "__description__": plugin.__description__,
                "__author__": plugin.__author__,
                "__version__": plugin.__version__,
            }

    def load_plugins(self):
        """Load all plugins"""
        for file in self.plugins():
            plugin = self.import_plugin(file)
            self._plugins.append(plugin)

    def start_plugins(self):
        """Start all plugins"""
        for plugin in self._plugins:
            plugin.load_settings()
            plugin.start()
            log.debug(f"{plugin.__alias__} was started")

    def stop_plugins(self):
        """Stop all plugins"""
        for plugin in self._plugins:
            plugin.stop()
            plugin.save_settings()
            log.debug(f"{plugin.__alias__} was stoped")

    @staticmethod
    def import_plugin(module: str) -> AbstractPlugin:
        """Import plugin by module name

        Args:
            module (str): Module name

        Returns:
            AbstractPlugin: Plugin instance
        """
        return importlib.import_module(module).load()



###############################################################################
# Decorator decorator is a simplified version of the code from the funcy lib.
# https://github.com/Suor/funcy
###############################################################################
from skippy.gui import CriticalMessageBox

from skippy.utils.logger import log

import functools
import traceback
import inspect


class Call:
    """A call object to pass as first argument to decorator.

    Call object is just a proxy for decorated function
    with call arguments saved in its attributes.
    """

    def __init__(self, func, args, kwargs):
        self.func, self.args, self.kwargs = func, args, kwargs

    def __call__(self):
        return self.func(*self.args, **self.kwargs)


def decorator(deco):
    """Transforms a flat wrapper into decorator."""
    spec = inspect.getargspec(deco)
    if len(spec.args) > 1 or spec.varargs or spec.keywords:

        @functools.wraps(deco)
        def _fab(*dargs, **dkwargs):
            return make_decorator(deco, *dargs, **dkwargs)

        return _fab
    else:
        return functools.wraps(deco)(make_decorator(deco))


def make_decorator(deco, *dargs, **dkwargs):
    def _decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            call = Call(func, args, kwargs)
            return deco(call, *dargs, **dkwargs)

        return wrapper

    return _decorator


@decorator
def critical(call):
    """Log error and show critical message."""
    try:
        return call()
    except Exception as error:
        log.error(error, exc_info=True)
        CriticalMessageBox(type(error).__name__, traceback.format_exc()).exec_()


@decorator
def ignore(call, error=Exception, value=None):
    """Ignore selected error and return value (if selected)."""
    try:
        return call()
    except error:
        return value


@decorator
def debug(call):
    log.debug(
        f"""Function {call.func.__name__}('{"', '".join(map(str, call.args))}{"', '".join([f'{str(kwarg)}={str(call.kwargs[kwarg])}' for kwarg in call.kwargs])}')"""
    )
    result = call()
    log.debug(f"Function {call.func.__name__}(...) --> {result}")

    return result


__all__ = ["critical", "ignore", "debug"]



from PyQt5 import QtWidgets, QtCore, QtGui

from skippy.core.scpclient import SCPClient

from skippy.gui import utils

from skippy.utils import filehandlers, translator
import skippy.config

from typing import Optional
import os


class LoginDialog(QtWidgets.QDialog):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super(LoginDialog, self).__init__(parent)
        self._layout = QtWidgets.QVBoxLayout(self)

        self.label = QtWidgets.QLabel(
            translator.Translator().translate("DIALOG.SIGN_IN_TO_WIKIDOT_LABEL"), self
        )
        self.login_box = QtWidgets.QLineEdit()
        self.login_box.setPlaceholderText(
            translator.Translator().translate("DIALOG.LOGIN_PLACEHOLDER")
        )
        self.password_box = QtWidgets.QLineEdit()
        self.password_box.setEchoMode(QtWidgets.QLineEdit.Password)
        self.password_box.setPlaceholderText(
            translator.Translator().translate("DIALOG.PASSWORD_PLACEHOLDER")
        )
        self.button = QtWidgets.QPushButton(
            translator.Translator().translate("DIALOG.OK_BUTTON"), self
        )

        self.button.clicked.connect(self.login)

        self._layout.addWidget(self.label, alignment=QtCore.Qt.AlignCenter)
        self._layout.addWidget(self.login_box)
        self._layout.addWidget(self.password_box)
        self._layout.addWidget(self.button, alignment=QtCore.Qt.AlignRight)

        self.setLayout(self._layout)

        self.setWindowTitle(f"Skippy - {skippy.config.version}")
        self.setWindowIcon(
            QtGui.QIcon(os.path.join(skippy.config.RESOURCES_FOLDER, "skippy.ico"))
        )
        self.move(300, 300)
        self.resize(200, 100)

        self.exec_()

    def login(self):
        login = self.login_box.text()
        password = self.password_box.text()

        SCPClient().auth(login, password)
        filehandlers.ProfileHandler().save(login, password)

        utils.getMainWindow().loginStatus.updateStatus(login)

        self.accept()



from PyQt5 import QtWidgets, QtCore, QtGui

from skippy.api import Action

from skippy.core.elements import elements as elems

from skippy.gui.dialogs import previewer, elements, finder
from skippy.gui import settings, utils

from skippy.utils.translator import Translator
import skippy.config

from typing import Optional, Union
from functools import partial
import os


class ActionBarBase:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.theme = settings.Settings().theme
        mainwindow = utils.getMainWindow()

        self.new_action = Action(
            Translator().translate("MENU_BAR.ACTION.FILE.NEW_NAME"),
            Translator().translate("MENU_BAR.ACTION.FILE.NEW_STATUS_TIP"),
            lambda: mainwindow.tab.newTab(),
            "new",
        )

        self.open_action = Action(
            Translator().translate("MENU_BAR.ACTION.FILE.OPEN_NAME"),
            Translator().translate("MENU_BAR.ACTION.FILE.OPEN_STATUS_TIP"),
            lambda: mainwindow.download(),
            "open",
        )

        self.upload_action = Action(
            Translator().translate("MENU_BAR.ACTION.FILE.UPLOAD_NAME"),
            Translator().translate("MENU_BAR.ACTION.FILE.UPLOAD_STATUS_TIP"),
            lambda: mainwindow.upload(mainwindow.tab.currentWidget().pdata),
            "upload",
        )

        self.upload_as_action = Action(
            Translator().translate("MENU_BAR.ACTION.FILE.UPLOAD_AS_NAME"),
            Translator().translate("MENU_BAR.ACTION.FILE.UPLOAD_AS_STATUS_TIP"),
            lambda: mainwindow.upload_as(mainwindow.tab.currentWidget().pdata),
            "upload_as",
        )

        self.close_action = Action(
            Translator().translate("MENU_BAR.ACTION.FILE.CLOSE_NAME"),
            Translator().translate("MENU_BAR.ACTION.FILE.CLOSE_STATUS_TIP"),
            lambda: mainwindow.tab.removeTab(mainwindow.tab.currentIndex()),
            "close",
        )

        self.load_files_action = Action(
            Translator().translate("MENU_BAR.ACTION.FILE.LOAD_FILES_NAME"),
            Translator().translate("MENU_BAR.ACTION.FILE.LOAD_FILES_STATUS_TIP"),
            lambda: mainwindow.load_files(),
            "load_files",
        )

        self.load_session_action = Action(
            Translator().translate("MENU_BAR.ACTION.FILE.LOAD_SESSION_NAME"),
            Translator().translate("MENU_BAR.ACTION.FILE.LOAD_SESSION_STATUS_TIP"),
            lambda: mainwindow.load_session(),
            "load_session",
        )

        self.save_session_action = Action(
            Translator().translate("MENU_BAR.ACTION.FILE.SAVE_SESSION_NAME"),
            Translator().translate("MENU_BAR.ACTION.FILE.SAVE_SESSION_STATUS_TIP"),
            lambda: mainwindow.save_session(),
            "save_session",
        )

        self.exit_action = Action(
            Translator().translate("MENU_BAR.ACTION.FILE.EXIT_NAME"),
            Translator().translate("MENU_BAR.ACTION.FILE.EXIT_STATUS_TIP"),
            lambda: mainwindow.close(),
            "exit",
        )

        self.undo_action = Action(
            Translator().translate("MENU_BAR.ACTION.EDIT.UNDO_NAME"),
            Translator().translate("MENU_BAR.ACTION.EDIT.UNDO_STATUS_TIP"),
            lambda: mainwindow.tab.currentWidget().editor.undo(),
            "undo",
        )

        self.redo_action = Action(
            Translator().translate("MENU_BAR.ACTION.EDIT.REDO_NAME"),
            Translator().translate("MENU_BAR.ACTION.EDIT.REDO_STATUS_TIP"),
            lambda: mainwindow.tab.currentWidget().editor.redo(),
            "redo",
        )

        self.cut_action = Action(
            Translator().translate("MENU_BAR.ACTION.EDIT.CUT_NAME"),
            Translator().translate("MENU_BAR.ACTION.EDIT.CUT_STATUS_TIP"),
            lambda: mainwindow.tab.currentWidget().editor.cut(),
            "cut",
        )

        self.copy_action = Action(
            Translator().translate("MENU_BAR.ACTION.EDIT.COPY_NAME"),
            Translator().translate("MENU_BAR.ACTION.EDIT.COPY_STATUS_TIP"),
            lambda: mainwindow.tab.currentWidget().editor.copy(),
            "copy",
        )

        self.paste_action = Action(
            Translator().translate("MENU_BAR.ACTION.EDIT.PASTE_NAME"),
            Translator().translate("MENU_BAR.ACTION.EDIT.PASTE_STATUS_TIP"),
            lambda: mainwindow.tab.currentWidget().editor.paste(),
            "paste",
        )

        self.select_all_action = Action(
            Translator().translate("MENU_BAR.ACTION.EDIT.SELECT_ALL_NAME"),
            Translator().translate("MENU_BAR.ACTION.EDIT.SELECT_ALL_STATUS_TIP"),
            lambda: mainwindow.tab.currentWidget().editor.selectAll(),
            "select_all",
        )

        self.find_action = Action(
            Translator().translate("MENU_BAR.ACTION.EDIT.FIND_NAME"),
            Translator().translate("MENU_BAR.ACTION.EDIT.FIND_STATUS_TIP"),
            lambda: finder.FinderDialog(mainwindow),
            "find",
        )

        self.preview_action = Action(
            Translator().translate("MENU_BAR.ACTION.EDIT.PREVIEW_NAME"),
            Translator().translate("MENU_BAR.ACTION.EDIT.PREVIEW_STATUS_TIP"),
            lambda: previewer.Previewer(
                mainwindow.tab.currentWidget().pdata, mainwindow
            ).deleteLater(),
            "preview",
        )

        self.toggle_theme_action = Action(
            Translator().translate("MENU_BAR.ACTION.SETTINGS.TOGGLE_THEME_NAME"),
            Translator().translate("MENU_BAR.ACTION.SETTINGS.TOGGLE_THEME_STATUS_TIP"),
            lambda: mainwindow.toggle_theme(),
            "toggle",
        )

        self.toggle_autocomplete_action = Action(
            Translator().translate("MENU_BAR.ACTION.SETTINGS.TOGGLE_AC_NAME"),
            Translator().translate("MENU_BAR.ACTION.SETTINGS.TOGGLE_AC_STATUS_TIP"),
            lambda: mainwindow.settings.setValue(
                "acEnabled",
                {"true": False, "false": True}[mainwindow.settings.acEnabled],
            ),
            "toggle",
        )

        self.elements_menu = QtWidgets.QMenu(
            Translator().translate("MENU_BAR.ACTION.EDIT.INSERT_MENU")
        )
        for elem in elems:
            element = elem()
            self.addAction(
                Action(
                    Translator().translate(element.__alias__),
                    Translator().translate(element.__description__),
                    partial(elements.ElementDialog, element, mainwindow),
                ),
                self.elements_menu,
            )

        self.login_action = Action(
            Translator().translate("MENU_BAR.ACTION.SETTINGS.LOGIN_NAME"),
            Translator().translate("MENU_BAR.ACTION.SETTINGS.LOGIN_STATUS_TIP"),
            lambda: mainwindow.login(),
            "login",
        )

        self.logout_action = Action(
            Translator().translate("MENU_BAR.ACTION.SETTINGS.LOGOUT_NAME"),
            Translator().translate("MENU_BAR.ACTION.SETTINGS.LOGOUT_STATUS_TIP"),
            lambda: mainwindow.logout(),
            "logout",
        )

        self.language_menu = QtWidgets.QMenu(
            Translator().translate("MENU_BAR.ACTION.SETTINGS.LANGUAGES_MENU")
        )
        for lang in Translator().languages():
            name = Translator().getLangName(lang)
            self.addAction(
                Action(name, name, partial(mainwindow.update_translate, lang)),
                self.language_menu,
            )

    def addAction(
        self, action: Action, menu: Optional[QtWidgets.QMenu] = None
    ) -> QtWidgets.QAction:
        qaction = QtWidgets.QAction(action.label, self)

        if action.img:
            qaction.setIcon(
                QtGui.QIcon(
                    os.path.join(
                        skippy.config.RESOURCES_FOLDER, self.theme, f"{action.img}.png"
                    )
                )
            )

        qaction.setStatusTip(action.statusTip)
        qaction.triggered.connect(action.func)

        if menu:
            menu.addAction(qaction)
        else:
            super().addAction(qaction)

        return qaction

    def addMenu(
        self, qmenu: Union[str, QtWidgets.QMenu], menu: Optional[QtWidgets.QMenu] = None
    ) -> QtWidgets.QMenu:
        if type(qmenu) == str:
            qmenu = QtWidgets.QMenu(qmenu, self)

        if menu:
            menu.addMenu(qmenu)
        else:
            super().addMenu(qmenu)

        return qmenu


class MenuBar(ActionBarBase, QtWidgets.QMenuBar):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super(MenuBar, self).__init__(parent)
        self.file_menu = self.addMenu(Translator().translate("MENU_BAR.FILE_MENU"))
        self.addAction(self.new_action, self.file_menu)
        self.addAction(self.open_action, self.file_menu)
        self.addAction(self.upload_action, self.file_menu)
        self.addAction(self.upload_as_action, self.file_menu)
        self.addAction(self.close_action, self.file_menu)
        self.file_menu.addSeparator()
        self.addAction(self.load_files_action, self.file_menu)
        self.file_menu.addSeparator()
        self.addAction(self.load_session_action, self.file_menu)
        self.addAction(self.save_session_action, self.file_menu)
        self.file_menu.addSeparator()
        self.addAction(self.exit_action, self.file_menu)

        self.edit_menu = self.addMenu(Translator().translate("MENU_BAR.EDIT_MENU"))
        self.addAction(self.undo_action, self.edit_menu)
        self.addAction(self.redo_action, self.edit_menu)
        self.edit_menu.addSeparator()
        self.addAction(self.cut_action, self.edit_menu)
        self.addAction(self.copy_action, self.edit_menu)
        self.addAction(self.paste_action, self.edit_menu)
        self.addAction(self.select_all_action, self.edit_menu)
        self.edit_menu.addSeparator()
        self.addAction(self.find_action, self.edit_menu)
        self.edit_menu.addSeparator()
        self.addAction(self.preview_action, self.edit_menu)
        self.addMenu(self.elements_menu, self.edit_menu)

        self.settings_menu = self.addMenu(
            Translator().translate("MENU_BAR.SETTINGS_MENU")
        )
        self.addAction(self.toggle_theme_action, self.settings_menu)
        self.addAction(self.toggle_autocomplete_action, self.settings_menu)
        self.settings_menu.addMenu(self.language_menu)
        self.settings_menu.addSeparator()
        self.addAction(self.login_action, self.settings_menu)
        self.addAction(self.logout_action, self.settings_menu)


class ToolBar(ActionBarBase, QtWidgets.QToolBar):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super(ToolBar, self).__init__(parent)
        self.addAction(self.new_action)
        self.addAction(self.open_action)
        self.addAction(self.upload_action)
        self.addAction(self.upload_as_action)
        self.addAction(self.close_action)
        self.addAction(self.load_files_action)
        self.addAction(self.undo_action)
        self.addAction(self.redo_action)
        self.addAction(self.cut_action)
        self.addAction(self.copy_action)
        self.addAction(self.paste_action)
        self.addAction(self.select_all_action)
        self.addAction(self.preview_action)

        self.setAllowedAreas(QtCore.Qt.LeftToolBarArea | QtCore.Qt.TopToolBarArea)


class ContextMenu(ActionBarBase, QtWidgets.QMenu):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super(ContextMenu, self).__init__(parent)
        self.addAction(self.new_action)
        self.addAction(self.load_files_action)
        self.addSeparator()
        self.addAction(self.undo_action)
        self.addAction(self.redo_action)
        self.addSeparator()
        self.addAction(self.cut_action)
        self.addAction(self.copy_action)
        self.addAction(self.paste_action)
        self.addAction(self.select_all_action)
        self.addSeparator()
        self.addAction(self.preview_action)
        self.addMenu(self.elements_menu)



from PyQt5 import QtCore

from typing import Any


class Settings(QtCore.QSettings):
    values = {
        "theme": "light",
        "lang": "en",
        "size": QtCore.QSize(700, 700),
        "pos": QtCore.QPoint(200, 200),
        "state": QtCore.Qt.WindowNoState,
        "toolbarArea": QtCore.Qt.LeftToolBarArea,
        "acEnabled": "true",
    }

    def __init__(self):
        super(Settings, self).__init__("skippy", "skippy")
        for key in self.allKeys():
            if key not in self.values:
                self.remove(key)

    def __getattr__(self, name: str) -> Any:
        if name in self.values:
            return self.value(name, self.values[name])
        raise AttributeError(
            f"{self.__class__.__name__} object has no attribute {name}"
        )

    def __setattr__(self, name: str, value: Any):
        self.setValue(name, value)



from abc import ABCMeta, abstractproperty
from typing import Any


class AbstractComponent(metaclass=ABCMeta):
    @abstractproperty
    def value(self) -> Any:
        pass

    @value.setter
    def value(self, value: Any):
        pass



"""Wikidot syntax previewer
"""
from skippy.api import PageData

from skippy.utils.logger import log

from requests.exceptions import RequestException
from abc import ABCMeta, abstractmethod
from typing import Union, Tuple, List
import unicodedata
import pyftml
import pyscp
import html
import re


def render(pdata: PageData) -> str:
    """Render page by page data

    Args:
        pdata (PageData): Page data (title, source, tags and files)

    Returns:
        str: Rendered HTML
    """
    preprocess = PreProcessorsHandler(pdata).process()
    htmlprocess = HTMLProcessorsHandler(preprocess, pdata).process()
    return PostProcessorsHandler(htmlprocess, pdata).process()


#################################################
# Base Classes
#################################################
class AbstractProcessor(metaclass=ABCMeta):

    """Abstract processor class"""

    pattern: str

    def __init__(self, source: str, pdata: PageData):
        """Initializing Processor

        Args:
            source (str): Page source
            pdata (PageData): Page data
        """
        self.source: str = source
        self.pdata: PageData = pdata

    @property
    def matches(self) -> List[Union[Tuple[str, ...], str]]:
        """Get all matches in source by pattern

        Returns:
            List[Union[Tuple[str, ...], str]]: List of matches
        """
        return re.findall(
            self.pattern,
            self.source,
        )

    @abstractmethod
    def process(self):
        """Abstract process method"""
        pass


class ProcessorsHandlerBase:

    """Abstract processors handler class"""

    def __init__(self, source: str, pdata: PageData):
        """Summary

        Args:
            source (str): Page source
            pdata (PageData): Page data
        """
        self.source: str = source
        self.pdata: PageData = pdata
        self.processors: List[AbstractProcessor] = []

    def register(self, processor: AbstractProcessor) -> object:
        """Register a processor

        Args:
            processor (AbstractProcessor): Processor class
        """
        self.processors.append(processor)

    def process(self) -> str:
        """Run all processor

        Returns:
            str: Processed source
        """
        for processor in self.processors:
            try:
                self.source = processor(self.source, self.pdata).process()
            except Exception as e:
                log.error(e, exc_info=True)
        return self.source


#################################################
# ProcessorsHandlers
#################################################
class PreProcessorsHandler(ProcessorsHandlerBase):

    """Handler of preprocessors"""

    def __init__(self, pdata: PageData):
        """Initializing preprocessors handler

        Args:
            pdata (PageData): Page data
        """
        super(PreProcessorsHandler, self).__init__(pdata["source"], pdata)
        self.register(IncludesProcessor)
        self.register(IftagsProcessor)


class HTMLProcessorsHandler(ProcessorsHandlerBase):

    """Handler of HTML processors"""

    def __init__(self, source: str, pdata: PageData):
        """Initializing HTML processors handler

        Args:
            source (str): Page source
            pdata (PageData): Page data
        """
        super(HTMLProcessorsHandler, self).__init__(source, pdata)
        self.register(MarkdownProcessor)
        self.register(InsertDataProcessor)
        self.register(ModuleCSSProcessor)

    def process(self) -> str:
        """Run all processor

        Returns:
            str: Processed source
        """
        self.html = self.processors[0](self.source, self.pdata).process()
        for processor in self.processors[1:]:
            try:
                self.html = processor(self.source, self.pdata, self.html).process()
            except Exception as e:
                log.error(e, exc_info=True)
        return self.html


class PostProcessorsHandler(ProcessorsHandlerBase):

    """Handler of postprocessors"""

    def __init__(self, source: str, pdata: PageData):
        """Initializing postprocessors handler

        Args:
            source (str): Page source
            pdata (PageData): Page data
        """
        super(PostProcessorsHandler, self).__init__(source, pdata)
        self.register(LocalImagesProcessor)
        self.register(HTMLTagsProcessor)


#################################################
# PreProcessors
#################################################
class IncludesProcessor(AbstractProcessor):

    """Include processor"""

    pattern: str = r"(\[\[include\s(?::.+?:|)(?:.+?:|)(?:.+)(?:\s((?:.|\n)+?)|)]])"

    def process(self, iteration: int = 0) -> str:
        """Replace all include tags with included page source

        Args:
            iteration (int, optional): Include iteration (max=5)

        Returns:
            str: Processed source
        """
        for include in self.matches:
            path = re.findall(
                r"(?::(.+?):|)((?:.+?:|).+)",
                include[0].split()[1].replace("]]", ""),
            )[0]
            site = path[0]
            page = path[1]
            try:
                wiki = pyscp.wikidot.Wiki(site)
                p = wiki(page).source
                args = [
                    i[1:] if i.startswith("\n") else i for i in include[1].split("|")
                ]
                if not args[0]:
                    args = []
                for arg in args:
                    arg = re.findall(
                        r"([\w-]+)(?:(?:\s|)=(?:\s|))((?:.|\n+?)+)",
                        arg,
                    )[0]
                    p = p.replace("{$" + arg[0] + "}", arg[1])
                self.source = self.source.replace(include[0], p)
            except Exception as e:
                log.error(e)
        if (
            re.findall(
                r"(\[\[include(?:.|\n)+?]])",
                self.source,
            )
            and iteration != 5
        ):
            self.process(iteration + 1)
        return self.source


class IftagsProcessor(AbstractProcessor):

    """Iftags processor"""

    pattern: str = r"(\[\[iftags (.+?)]]((?:(?:.|\n|\s)+?))\[\[\/iftags]])"

    def process(self) -> str:
        """Remove iftags if it not matches with page tags

        Returns:
            str: Processed source
        """
        for iftag in self.matches:
            match = True
            elem = re.findall(r"(\+.+?(?=\s|$)|-.+?(?=\s|$))", iftag[1])
            for e in elem:
                if e.startswith("+"):
                    if e[1:] not in self.pdata["tags"]:
                        match = False
                elif e[1:] in self.pdata["tags"]:
                    match = False
            if match:
                self.source = self.source.replace(iftag[0], iftag[2])
        return self.source


#################################################
# HTMLProcessors
#################################################
class MarkdownProcessor(AbstractProcessor):

    """Markdown processor"""

    wiki: pyscp.wikidot.Wiki = pyscp.wikidot.Wiki("www.wikidot.com")

    def _wikidot(self) -> str:
        """If connected to internet, get previewed HTML from Wikidot

        Returns:
            str: Previewed source
        """
        return self.wiki._module("edit/PagePreviewModule", source=self.source)["body"]

    def _ftml(self) -> str:
        """If don't have connection to internet, get previewed HTML from local FTML previewer

        Returns:
            str: Previewed source
        """
        return pyftml.render_html(self.source)["body"]

    def process(self) -> str:
        """Get previewed page with online/offline methods

        Returns:
            str: Processed source
        """
        try:
            return self._wikidot()
        except RequestException:
            return self._ftml()


class InsertDataProcessor(AbstractProcessor):

    """InsertData processor"""

    html_base: str = """
        <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
        <html xmlns="http://www.w3.org/1999/xhtml" xml:lang="ru" lang="ru" pdata-lt-installed="true">
            <head>
                <style type="text/css">
                    @import url(http://d3g0gp89917ko0.cloudfront.net/v--3e3a6f7dbcc9/common--theme/base/css/style.css);
                    @import url(http://scp-ru.wdfiles.com/local--code/component:theme2/1);
                </style>
                <script type="text/javascript" src="http://d3g0gp89917ko0.cloudfront.net/v--3e3a6f7dbcc9/common--javascript/init.combined.js"></script>
                <script type="text/javascript">
                    var URL_HOST = 'www.wikidot.com';
                    var URL_DOMAIN = 'wikidot.com';
                    var USE_SSL =  true ;
                    var URL_STATIC = 'http://d3g0gp89917ko0.cloudfront.net/v--3e3a6f7dbcc9';
                    // global request information
                    
                    var WIKIREQUEST = {};
                    WIKIREQUEST.info = {};
                    
                    WIKIREQUEST.info.domain = "www.wikidot.com";
                    WIKIREQUEST.info.siteId = 648902;
                    WIKIREQUEST.info.siteUnixName = "www";
                    WIKIREQUEST.info.categoryId = 4388020;
                    WIKIREQUEST.info.themeId = 192064;
                    WIKIREQUEST.info.requestPageName = "start:start";
                    OZONE.request.timestamp = 1619763822;
                    OZONE.request.date = new Date();
                    WIKIREQUEST.info.lang = 'en';
                            WIKIREQUEST.info.pageUnixName = "start:start";
                    WIKIREQUEST.info.pageId = 22129557;
                                    WIKIREQUEST.info.lang = "en";
                    OZONE.lang = "en";
                    var isUAMobile = !!/Android|webOS|iPhone|iPod|BlackBerry|IEMobile|Opera Mini/i.test(navigator.userAgent);
                </script>
                <script type="text/javascript">
            
                    require.config({
                        baseUrl: URL_STATIC + '/common--javascript',
                        paths: {
                            'jquery.ui': 'jquery-ui.min',
                            'jquery.form': 'jquery.form'
                        }
                    });
            
                </script>
                <script type="text/javascript" src="http://d3g0gp89917ko0.cloudfront.net/v--3e3a6f7dbcc9/common--javascript/WIKIDOT.combined.js"></script>
                <<MODULE-CSS-PREVIEW>>
                <style>
                    body {
                        padding: 2em !important;
                    }
                </style>
            </head>
            <body id="html-body">
                <div id="dummy-ondomready-block" style="display:none;"></div>
                <div id="page-title">
                    <<TITLE>>
                </div>
                <div id="page-content" style="min-height: 500px;">
                    <<CONTENT>>
                </div>
                <div class="page-tags">
                    <span>
                        <<TAGS>>
                    </span>
                </div>
            </body>
        </html>
        """

    def __init__(self, source: str, pdata: PageData, html: str):
        """Initializing InsertData processor

        Args:
            source (str): Page source
            pdata (PageData): Page data
            html (str): Previewed page html
        """
        super(InsertDataProcessor, self).__init__(source, pdata)
        self.html: str = html

    def process(self) -> str:
        """Insert page data to HTML template

        Returns:
            str: Processed HTML page
        """
        self.source = self.html_base.replace(
            "<<TITLE>>", html.escape(self.pdata["title"])
        )
        self.source = self.source.replace("<<CONTENT>>", self.html)
        self.source = self.source.replace(
            "<<TAGS>>",
            " ".join([f"<a href='#'>{tag}</a>" for tag in self.pdata["tags"]]),
        )
        return self.source


class ModuleCSSProcessor(AbstractProcessor):

    """Module CSS processor"""

    pattern: str = r"(?:(?:\[\[module) (?:CSS|css)(?:(.+?)|)\]\]\n)((.|\n)+?)(?:\n\[\[\/module\]\])"

    def __init__(self, source: str, pdata: PageData, html: str):
        """Initializing Module CSS processor

        Args:
            source (str): Page source
            pdata (PageData): Page data
            html (str): Previewed page html
        """
        super(ModuleCSSProcessor, self).__init__(source, pdata)
        self.html: str = html

    def process(self) -> str:
        """Convert [[module CSS]] block to <style> tag

        Returns:
            str: Processed HTML page
        """
        styles = ""
        for style in self.matches:
            style = unicodedata.normalize("NFKD", style[1])
            styles += f"<style>\n{style}\n</style>\n"
        self.html = self.html.replace("<<MODULE-CSS-PREVIEW>>", styles)
        return self.html


#################################################
# PostProcessor
#################################################
class LocalImagesProcessor(AbstractProcessor):

    """Local images processor"""

    pattern: str = r'<img src="http://www.wdfiles.com/local--files//(.+?)"'

    def process(self) -> str:
        """Insert local images to page

        Returns:
            str: Processed HTML page
        """
        files = self.pdata["files"]
        for img in self.matches:
            if img in files:
                self.source = self.source.replace(
                    f"http://www.wdfiles.com/local--files//{img}",
                    f"data:{img.split('.')[1]}/;base64,{files[img]}",
                )
        return self.source


class HTMLTagsProcessor(AbstractProcessor):

    """HTML tags processor"""

    pattern: str = r"(\[\[html(?:(?:.+?)| |)]]((?:.|\n|\s)+?)\[\[\/html]])"

    def process(self) -> str:
        """Unescape [[html]] tags

        Returns:
            str: Processed HTML page
        """
        for tag in self.matches:
            self.source = self.source.replace(tag[0], html.unescape(tag[1]))
        return self.source



from PyQt5 import QtWidgets

from skippy.api import critical

from skippy.core import autoupdate, scpclient

from skippy.gui.dialogs import login, updater
from skippy.gui.mainwindow import Skippy

from skippy.utils import filehandlers, logger

import sys


@critical
def start_ui():
    """Start Skippy application."""
    logger.log.info("Skippy was started...")

    scpclient.SCPClient(*filehandlers.ProfileHandler().load())

    exit_code = Skippy.EXIT_CODE_REBOOT
    while exit_code == Skippy.EXIT_CODE_REBOOT:
        app = QtWidgets.QApplication(sys.argv)
        app.setApplicationName("skippy")

        window = Skippy()

        result = True
        updateClient = autoupdate.AbstractUpdateClient.getClient()
        if updateClient.checkVersion():
            result = updater.UpdaterDialog(updateClient.version, updateClient.update)

        if not filehandlers.ProfileHandler().load()[0]:
            result = login.LoginDialog().result()

        if not result:
            break

        window.show()

        exit_code = app.exec_()
        app = None
    logger.log.info("Skippy was stopped...")

    return exit_code



from PyQt5 import QtWidgets, QtCore

from typing import Optional

ENTER_KEYS = (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter)
DELETE_KEYS = (QtCore.Qt.Key_Backspace, QtCore.Qt.Key_Delete)


def getApplication() -> Optional[QtCore.QCoreApplication]:
    return QtWidgets.QApplication.instance()


def getMainWindow() -> Optional[QtWidgets.QMainWindow]:
    app = getApplication()
    for widget in app.topLevelWidgets():
        if isinstance(widget, QtWidgets.QMainWindow):
            return widget
    return None


def showStatusMessage(message: str):
    getMainWindow().statusBar().showMessage(message)



from PyQt5 import QtCore


class AbstractWorker(QtCore.QObject):
    finished = QtCore.pyqtSignal()

    def run(self):
        raise NotImplementedError


class Thread(QtCore.QThread):
    def __init__(self, worker: AbstractWorker):
        super(Thread, self).__init__()
        self.worker = worker
        self.worker.moveToThread(self)

        self.worker.finished.connect(self.quit)
        self.worker.finished.connect(self.worker.deleteLater)

        self.started.connect(self.worker.run)
        self.finished.connect(self.deleteLater)



from skippy.gui import components

from abc import ABCMeta, abstractmethod, abstractproperty
from typing import Any


class AbstractSetting(metaclass=ABCMeta):
    __alias__: str
    __description__: str

    _component: components.AbstractComponent

    def __init__(self):
        self.component = self._component()

    @abstractproperty
    def value(self) -> Any:
        pass

    @value.setter
    def value(self, value: Any):
        pass

    @abstractmethod
    def from_toml(self, text: str):
        pass

    @abstractmethod
    def to_toml(self) -> str:
        pass



from PyQt5 import QtCore

from skippy.api import PageData

from skippy.core import scpclient

from skippy.gui import thread

from skippy.utils import discord_rpc

from typing import List
import os


class FileWorker(thread.AbstractWorker):
    progress = QtCore.pyqtSignal(str, bytes)

    def __init__(self, files: List[str]):
        super(FileWorker, self).__init__()
        self.files = files

    def run(self):
        for file in self.files:
            if file:
                with open(file, "rb") as f:
                    self.progress.emit(os.path.split(file)[1], f.read())
        self.finished.emit()


class UploadWorker(thread.AbstractWorker):
    def __init__(self, pdata: PageData, comment: str = "Edit using Skippy"):
        super(UploadWorker, self).__init__()
        self.pdata = pdata
        self.comment = comment

    def run(self):
        scpclient.SCPClient().upload(self.pdata, self.comment)
        self.finished.emit()


class DiscordRPCWorker(thread.AbstractWorker):
    def __init__(self, title: str, words: int, letters: int):
        super(DiscordRPCWorker, self).__init__()
        self.rpc = discord_rpc.DiscordRPC()

        self.title = title
        self.words = words
        self.letters = letters

    def run(self):
        self.rpc.update(self.title, self.words, self.letters)

        self.finished.emit()



from typing import (
    MutableMapping,
    NamedTuple,
    TypedDict,
    Callable,
    Optional,
    Tuple,
    Dict,
    List,
    Any,
)


class Language(NamedTuple):
    code: str
    dictionary: MutableMapping[str, Any]


class Field(NamedTuple):
    tag: str
    name: str
    description: str


class Action(NamedTuple):
    label: str
    statusTip: str
    func: Callable[..., Any]
    img: Optional[str] = None


class InlineSyntax(NamedTuple):
    opening: str
    closing: str
    multiple: bool


class BlockSyntax(NamedTuple):
    closeBlock: bool
    inline: bool
    dataVar: Dict[str, Dict[str, str]] = {"": {}}


class PageData(TypedDict):
    title: str
    source: str
    tags: List[str]
    files: Dict[str, str]
    link: Optional[Tuple[str, str]]


__all__ = ["Language", "Field", "Action", "InlineSyntax", "BlockSyntax", "PageData"]



"""Discord Rich Presence
"""
from skippy.api import Singleton, ignore

from pypresence import PyPresenceException, Presence
from struct import error as StructError
import time


class DiscordRPC(Presence, metaclass=Singleton):

    """Discord Rich Presence class"""

    def __init__(self):
        """Init DiscordRPC"""
        super(DiscordRPC, self).__init__("828818867568640071")

        self._words = 0
        self._letters = 0
        self._time = int(time.mktime(time.localtime()))

    @ignore((PyPresenceException, StructError))
    def connect(self):
        """Connect to Discord RPC"""
        super(DiscordRPC, self).connect()
        self.update("SCP")

    @ignore((PyPresenceException, AssertionError))
    def update(self, title: str, words: int = 0, letters: int = 0):
        super(DiscordRPC, self).update(
            state=f"Words: {str(words)}, Letters: {str(letters)}",
            details=f'Writing an "{title}"',
            start=self._time,
            small_text="Skippy",
            small_image="none",
            large_text="Skippy",
            large_image="skippy",
        )

    @ignore((PyPresenceException, AssertionError))
    def close(self):
        """Close RPC connection"""
        super(DiscordRPC, self).close()



"""Singleton facade for pyscp module.
"""
from skippy.api import Singleton, PageData, critical, ignore

from urllib.parse import urlparse
from pyscp.wikidot import Wiki
from typing import Optional
import requests
import base64


RequestException = requests.exceptions.RequestException


class SCPClient(metaclass=Singleton):

    """Singleton facade for pyscp module."""

    def __init__(self, login: Optional[str] = None, password: Optional[str] = None):
        """Initialize SCPClient.

        Args:
            login (Optional[str], optional): User login
            password (Optional[str], optional): User password
        """
        self._session = None

        if login and password:
            self.auth(login, password)

    @critical
    @ignore(RequestException)
    def auth(self, login: str, password: str):
        """Retrieving Wikidot Session from Credentials.

        Args:
            login (str): User login
            password (str): User password
        """
        wikidot = Wiki("www.wikidot.com")
        wikidot.auth(login, password)

        self._session = wikidot.cookies

    @critical
    @ignore(RequestException)
    def get_wiki(self, site: str) -> Wiki:
        """Get Wiki object instance by Wikidot site name with current session cookies.

        Args:
            site (str): Wikidot site name

        Returns:
            Wiki: Wiki object instance with current session cookies
        """
        wiki = Wiki(site)
        if self._session:
            wiki.cookies = self._session

        return wiki

    @staticmethod
    def download_file(url: str):
        with requests.get(url) as res:
            return base64.b64encode(res.content).decode("utf-8")

    @critical
    @ignore(RequestException)
    def upload(self, page: PageData, comment: str = "Edit using Skippy"):
        """Upload page with tags and files to selected Wikidot site.

        Args:
            page (PageData): Page for given Wikidot site
            comment (str, optional): Comment for uploaded changes
        """
        wiki = self.get_wiki(page["link"][0])
        p = wiki(page["link"][1])

        try:
            p.edit(title=page["title"], source=page["source"], comment=comment)
        except AttributeError:
            p.create(title=page["title"], source=page["source"], comment=comment)

        p.set_tags(page["tags"])

        file_urls = {file.name: file.url for file in p.files}

        for file in page["files"]:
            file_source = base64.b64decode(page["files"][file])
            if file not in file_urls:
                p.upload(file, file_source)
            elif file in file_urls and page["files"][file] != self.download_file(
                file_urls[file]
            ):
                p.remove_file(file)
                p.upload(file, file_source)

        for file in p.files:
            name = file.name
            if name not in page["files"]:
                p.remove_file(name)

    @ignore(Exception, {})
    @ignore(RequestException)
    def download(self, site: str, page: str) -> PageData:
        """Download page with tags and files fron selected Wikidot site.

        Args:
            site (str): Wikidot site name
            page (str): Page for given Wikidot site

        Returns:
            PageData: Page data
        """
        wiki = self.get_wiki(site)
        p = wiki(page)

        files = {file.name: self.download_file(file.url) for file in p.files}
        return {
            "title": p.title,
            "source": p.source,
            "tags": list(p.tags),
            "files": files,
            "link": (urlparse(wiki.site).netloc, page),
        }



from PyQt5 import QtWidgets, QtGui

import skippy.config

import os


class CriticalMessageBox(QtWidgets.QDialog):
    def __init__(self, error: str, traceback: str):
        super(CriticalMessageBox, self).__init__()
        self._layout = QtWidgets.QVBoxLayout(self)

        self.errorMessage = QtWidgets.QPlainTextEdit(self)
        self.errorMessage.setReadOnly(True)
        self.errorMessage.setPlainText(traceback)

        self._layout.addWidget(self.errorMessage)

        self.setLayout(self._layout)

        self.setWindowTitle(f"Error: {error}")
        self.setWindowIcon(
            QtGui.QIcon(os.path.join(skippy.config.RESOURCES_FOLDER, "skippy.ico"))
        )

        self.resize(530, 390)



from PyQt5 import QtWidgets, QtCore, QtGui

from skippy.core import scpclient

from skippy.gui import sitebox, thread, utils

from skippy.utils import translator
import skippy.config

from typing import Optional
import os


class DownloadWorker(thread.AbstractWorker):
    finished = QtCore.pyqtSignal(dict)

    def __init__(self, site: str, page: str):
        super(DownloadWorker, self).__init__()
        self.site = site
        self.page = page

    def run(self):
        pageData = scpclient.SCPClient().download(self.site, self.page)
        self.finished.emit(pageData)


class DownloadDialog(QtWidgets.QDialog):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super(DownloadDialog, self).__init__(parent)
        self._layout = QtWidgets.QVBoxLayout(self)

        self.label = QtWidgets.QLabel(
            translator.Translator().translate("DIALOG.ENTER_PAGE_LABEL"), self
        )

        self.site_box = sitebox.SiteBox(self)

        self.page_box = QtWidgets.QLineEdit()
        self.page_box.setPlaceholderText(
            translator.Translator().translate("DIALOG.PAGE_BOX_PLACEHOLDER")
        )

        self.button = QtWidgets.QPushButton(
            translator.Translator().translate("DIALOG.OK_BUTTON"), self
        )
        self.button.clicked.connect(self.download)

        self._layout.addWidget(self.label, alignment=QtCore.Qt.AlignCenter)
        self._layout.addWidget(self.site_box)
        self._layout.addWidget(self.page_box)
        self._layout.addWidget(self.button, alignment=QtCore.Qt.AlignRight)

        self.setLayout(self._layout)

        self.setWindowTitle(f"Skippy - {skippy.config.version}")
        self.setWindowIcon(
            QtGui.QIcon(os.path.join(skippy.config.RESOURCES_FOLDER, "skippy.ico"))
        )
        self.move(300, 300)
        self.resize(200, 100)

        self.show()

    def download(self):
        site = self.site_box.currentText()
        page = self.page_box.text()

        self._thread = thread.Thread(DownloadWorker(site, page))
        self._thread.worker.finished.connect(
            lambda page_data: utils.getMainWindow().tab.newTab(*page_data.values())
        )
        self._thread.start()

        self.close()



from PyQt5 import QtWidgets, QtCore

from skippy.api import ignore

from skippy.gui import thread

from skippy.utils import cached_property, translator, filehandlers

from requests.exceptions import RequestException
from typing import Optional, Callable, List, Any
from urllib.parse import urlparse
import pyscp


class UserSitesWorker(thread.AbstractWorker):
    finished = QtCore.pyqtSignal(list)

    def run(self):
        self.finished.emit(self.sites)

    @cached_property
    @ignore(RequestException, [])
    def sites(self) -> List[str]:
        return [
            urlparse(wiki.site).netloc
            for wiki in pyscp.wikidot.User(
                filehandlers.ProfileHandler().load()[0]
            ).member
        ]


class SiteBox(QtWidgets.QComboBox):
    def __init__(self, parent: Optional[QtWidgets.QWidget] = None, default: str = ""):
        super(SiteBox, self).__init__(parent)
        self.default = default

        self._lineEdit = QtWidgets.QLineEdit(self)
        self._lineEdit.setPlaceholderText(
            translator.Translator().translate("DIALOG.SITE_BOX_LINEEDIT")
        )

        self.setLineEdit(self._lineEdit)
        self.setEditable(True)

        self.setCurrentText(default)

        self._thread = thread.Thread(UserSitesWorker())
        self._thread.worker.finished.connect(self.addSites)
        self._thread.start()

    def addSites(self, sites: List[str]):
        self.addItems(sites)
        self.setCurrentText(self.default)
        self.setCompleter(QtWidgets.QCompleter(sites))
        if not self.currentText():
            self._lineEdit.clear()



from PyQt5 import QtWidgets, QtCore

from skippy.api import PageData, ignore

from skippy.gui import settings, resources, editor, utils
from skippy.gui.dialogs import filesdialog

from skippy.utils import translator, filehandlers

from typing import Optional, List, Dict, Tuple, Any
import base64


class ProjectList(QtWidgets.QTabWidget):
    titleChanged = QtCore.pyqtSignal(str)
    statsChanged = QtCore.pyqtSignal(str, int, int)

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super(ProjectList, self).__init__(parent)
        resources.qInitResources()

        self.tabCloseRequested.connect(self.removeTab)
        self.currentChanged.connect(lambda: self.titleChanged.emit(self.currentTitle()))
        self.currentChanged.connect(
            lambda: self.currentWidget().statusBarStats()
            if self.currentWidget()
            else None
        )
        self.titleChanged.connect(
            lambda title: self.statsChanged.emit(title, *self.currentEditorStats())
        )

        self.setStyleSheet(
            "QTabBar{font-family: Arial; font-size:10pt;} QTabBar::close-button {image: url(:"
            + settings.Settings().theme
            + "/close.png);}"
        )
        self.setTabsClosable(True)
        self.setMovable(True)
        self.setAcceptDrops(True)

    def newTab(
        self,
        title: str = "New tab",
        source: str = "",
        tags: List[str] = [],
        files: Dict[str, str] = {},
        link: Optional[Tuple[str, str]] = None,
    ):
        tab = TabWidget(title, source, tags, files, link, self)
        tab.titleChanged.connect(self.setTitle)
        tab.titleChanged.connect(self.titleChanged.emit)
        tab.sourceChanged.connect(self.save)
        tab.sourceChanged.connect(
            lambda: self.statsChanged.emit(
                self.currentTitle(), *self.currentEditorStats()
            )
        )

        self.addTab(tab, title)
        self.setCurrentIndex(self.count() - 1)

        self.save()

    def currentTitle(self) -> str:
        return self.tabText(self.currentIndex())

    @ignore(AttributeError, (0, 0))
    def currentEditorStats(self) -> Tuple[int, int]:
        return self.currentWidget().editorStats()

    def setTitle(self, text: str):
        self.setTabText(self.currentIndex(), text)

    def removeTab(self, index: int):
        super().removeTab(index)
        self.checkCount()

    def checkCount(self):
        if not self.count():
            self.newTab()

    def save(self, path: Optional[str] = None):
        session = {"session": []}
        for i in range(self.count()):
            widget = self.widget(i)
            if type(widget.pdata["tags"]) == str:
                widget.pdata["tags"] = widget.pdata["tags"].split()
            session["session"].append(widget.pdata)
        session["pos"] = self.currentIndex()

        filehandlers.SessionHandler(path).save(session)

    def load(self, path: Optional[str] = None):
        session = filehandlers.SessionHandler(path).load()

        self.clear()
        for page in session["session"]:
            self.newTab(
                page["title"],
                page["source"],
                page["tags"],
                page["files"],
                page["link"] if "link" in page else page["parent"],
            )
        self.checkCount()
        self.setCurrentIndex(session["pos"] if "pos" in session else self.count() - 1)

        self.titleChanged.emit(self.currentTitle())

        self.save()


class TabWidget(QtWidgets.QWidget):
    titleChanged = QtCore.pyqtSignal(str)
    sourceChanged = QtCore.pyqtSignal()

    def __init__(
        self,
        title: str,
        source: str,
        tags: List[str],
        files: Dict[str, str],
        link: Optional[Tuple[str, str]],
        parent: Optional[QtWidgets.QWidget] = None,
    ):
        super(TabWidget, self).__init__(parent)
        self._layout = QtWidgets.QVBoxLayout(self)

        self.pdata: PageData = {
            "title": title,
            "source": source,
            "tags": tags,
            "files": files,
            "link": link,
        }

        self.title_box = QtWidgets.QLineEdit()
        self.title_box.setText(title)
        self.title_box.textEdited.connect(self.titleChanged.emit)
        self.title_box.textEdited.connect(lambda text: self.setData("title", text))

        self.editor = editor.AdvancedEditor(self)
        self.editor.setPlainText(source)
        self.editor.textChanged.connect(self.statusBarStats)
        self.editor.textChanged.connect(
            lambda: self.setData("source", self.editor.toPlainText())
        )
        self.editor.fileDragAndDroped.connect(self.uploadFile)

        self.tags_box = QtWidgets.QLineEdit()
        self.tags_box.setText(" ".join(tags))
        self.tags_box.textChanged.connect(
            lambda text: self.setData("tags", text.split())
        )

        self.files_button = QtWidgets.QPushButton(
            translator.Translator().translate("MAIN.FILES_BUTTON"), self
        )
        self.files_button.clicked.connect(
            lambda: filesdialog.FilesDialog(self.pdata["files"], self)
        )

        self._layout.addWidget(self.title_box)
        self._layout.addWidget(self.editor)
        self._layout.addWidget(self.tags_box)
        self._layout.addWidget(self.files_button)
        self.setLayout(self._layout)

        self.setStyleSheet(
            "QLineEdit, QPlainTextEdit {font-family: Arial; font-size:12pt;}"
        )

    def editorStats(self) -> Tuple[int, int]:
        content = self.editor.toPlainText()
        return len(content.split()), len(content)

    def setData(self, param: str, data: Any):
        self.pdata[param] = data
        self.sourceChanged.emit()

    def uploadFile(self, filename: str, source: bytes):
        self.pdata["files"][filename] = base64.b64encode(source).decode("utf-8")

    def statusBarStats(self):
        words, letters = self.editorStats()
        utils.showStatusMessage(f"Words: {words}, Letters: {letters}")



"""Various file handlers.
"""
from skippy.utils import is_frozen

import skippy.config

from abc import ABCMeta, abstractmethod
from typing import Optional, Tuple, Dict, Any
import json
import os


class AbstractFileHandler(metaclass=ABCMeta):

    """Abstract file handler

    Attributes:
        filepath (str): Path to handle file
    """

    _file: str

    def __init__(self, filepath: Optional[str] = None):
        """Init FileHandler

        Args:
            filepath (Optional[str], optional): Path to handle file
        """
        self.filepath = filepath or os.path.join(
            skippy.config.PROPERTY_FOLDER, self._file
        )

    def read(self) -> str:
        """Read file

        Returns:
            str: File text
        """
        with open(self.filepath, "r") as file:
            return file.read()

    def write(self, text: str):
        """Write file

        Args:
            text (str): File text
        """
        with open(self.filepath, "w") as file:
            file.write(text)

    @abstractmethod
    def load(self):
        """Abstract load method"""
        pass

    @abstractmethod
    def save(self, *args, **kwargs):
        """Abstract save method"""
        pass


class ProfileHandler(AbstractFileHandler):

    """Profile handler

    Attributes:
        _file (str): Handled file name
    """

    _file = "profile.json"

    def load(self) -> Tuple[str, str]:
        """Load profile

        Returns:
            Tuple[str, str]: Tuple of user login and password
        """
        if os.path.exists(self.filepath) and self.read():
            profile = json.loads(self.read())
            return profile["login"], profile["password"]
        return "", ""

    def save(self, login: str, password: str):
        """Save profile

        Args:
            login (str): User login
            password (str): User password
        """
        data = {"login": login, "password": password}
        profile = json.dumps(data)

        self.write(profile)

    def logout(self):
        """Logout from account"""
        self.save("", "")


class SessionHandler(AbstractFileHandler):

    """Session Handler

    Attributes:
        _file (str): Handled file name
    """

    _file = "session.json"

    def load(self) -> Dict[str, Any]:
        """Load session

        Returns:
            Dict[str, Any]: Session
        """
        if os.path.exists(self.filepath) and self.read():
            return json.loads(self.read())
        return {"session": [], "pos": 0}

    def save(self, session: Dict[str, Any]):
        """Save session

        Args:
            session (Dict[str, Any]): Save session
        """
        self.write(json.dumps(session))



"""Example plugin
"""
from skippy.core.plugins import AbstractPlugin
from skippy.utils.logger import log


class ExamplePlugin(AbstractPlugin):

    """This is a Example plugin class"""

    __alias__ = "ExamplePlugin"

    __description__ = "I am a example plugin."
    __author__ = "MrNereof"
    __version__ = "1.0.0"

    def start(self):
        """do something on start"""
        log.debug("Start")

    def stop(self):
        """do something on stop"""
        log.debug("Stop")


def load() -> ExamplePlugin:
    """Load ExamplePlugin class

    Returns:
        ExamplePlugin: A ExamplePlugin instance
    """
    return ExamplePlugin()



"""The main window of Skippy.
"""
from PyQt5 import QtWidgets, QtGui, QtCore

from skippy.api import PageData, critical

from skippy.gui.dialogs import download, upload, login
from skippy.gui import (
    actionbar,
    tabwidget,
    loginstatus,
    workers,
    thread,
    styles,
    settings,
    utils,
)

from skippy.utils import translator, filehandlers

import skippy.config

import os


class Skippy(QtWidgets.QMainWindow):

    """The main window of Skippy."""

    EXIT_CODE_REBOOT = -123

    def __init__(self):
        """Create a new main window."""
        super(Skippy, self).__init__()
        self._threads = []

        self.settings = settings.Settings()

        translator.Translator().load(self.settings.lang)

        self.menuBar = actionbar.MenuBar(self)
        self.toolBar = actionbar.ToolBar(self)
        self.setMenuBar(self.menuBar)
        self.addToolBar(self.settings.toolbarArea, self.toolBar)

        self.tab = tabwidget.ProjectList(self)
        self.tab.statsChanged.connect(self.update_rpc)
        self.tab.titleChanged.connect(self.update_title)
        self.tab.load()

        self.loginStatus = loginstatus.LoginStatus(self)

        self.setCentralWidget(self.tab)

        self.status = QtWidgets.QStatusBar(self)
        self.setStatusBar(self.status)

        self.setWindowIcon(
            QtGui.QIcon(os.path.join(skippy.config.RESOURCES_FOLDER, "skippy.ico"))
        )
        self.resize(self.settings.size)
        self.move(self.settings.pos)
        self.setWindowState(QtCore.Qt.WindowState(self.settings.state))

        self.setFont(QtGui.QFont("Arial", 10))

        if self.settings.theme == "light":
            styles.light()
        else:
            styles.dark()

    def update_title(self, title: str):
        """Update title for Skippy window."""
        self.setWindowTitle(f"{title} | Skippy - {skippy.config.version}")

    def download(self):
        """Run page download dialog."""
        download.DownloadDialog(self)

    @critical
    def upload(self, pdata: PageData):
        """Upload page at prepared URL from parent field, if don't has it run upload dialog.

        Args:
            pdata (PageData): Page data
        """
        if pdata["link"]:
            uploadPageThread = thread.Thread(workers.UploadWorker(pdata))
            uploadPageThread.finished.connect(
                lambda: self.removeThread(uploadPageThread)
            )
            self._threads.append(uploadPageThread)
            uploadPageThread.start()
        else:
            self.upload_as(pdata)

    def upload_as(self, pdata: PageData):
        """Run upload dialog.

        Args:
            pdata (PageData): Page data
        """
        upload.UploadDialog(pdata, self)

    def login(self):
        """Run login dialog."""
        login.LoginDialog()

    def logout(self):
        """Logout from account and run a login dialog."""
        filehandlers.ProfileHandler().logout()
        self.hide()
        self.login()
        self.show()

    def save_session(self):
        """Save current work session to selected file."""
        path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, "Save file", "session.json", "JSON file (*.json)\nAll files (*.*)"
        )
        if path:
            self.tab.save(path)

    def load_session(self):
        """Load session from selected file."""
        path, _ = QtWidgets.QFileDialog.getOpenFileName(
            self, "Load file", "", "JSON file (*.json)\nAll files (*.*)"
        )
        if path:
            self.tab.load(path)

    def load_files(self):
        """Load files to the current page."""
        files, _ = QtWidgets.QFileDialog.getOpenFileNames(
            self, "Load files", "", "All files (*.*)"
        )
        loadFileThread = thread.Thread(workers.FileWorker(files))
        loadFileThread.finished.connect(lambda: self.removeThread(loadFileThread))
        self._threads.append(loadFileThread)
        loadFileThread.worker.progress.connect(self.tab.currentWidget().uploadFile)
        loadFileThread.start()

    def update_rpc(self, title: str, words: int, letters: int):
        for thr in self._threads:
            if type(thr.worker) == workers.DiscordRPCWorker:
                return
        updateRPCThread = thread.Thread(workers.DiscordRPCWorker(title, words, letters))
        updateRPCThread.finished.connect(lambda: self.removeThread(updateRPCThread))
        self._threads.append(updateRPCThread)
        updateRPCThread.start()

    def toggle_theme(self):
        """Toggle current theme."""
        self.settings.theme = "dark" if self.settings.theme == "light" else "light"
        self.restart()

    def update_translate(self, lang: str):
        """Update translate language.

        Args:
            lang (str): Language code
        """
        self.settings.lang = lang
        self.restart()

    @classmethod
    def restart(cls):
        """Restart window."""
        utils.getApplication().exit(cls.EXIT_CODE_REBOOT)

    def removeThread(self, thr: thread.Thread):
        if thr in self._threads:
            self._threads.remove(thr)

    def resizeEvent(self, event: QtGui.QResizeEvent):
        """Change size of login status widget when window resized.

        Args:
            event (QtGui.QResizeEvent): Resize event
        """
        self.loginStatus.move(self.width() - 200, -5)
        if self.width() < 350:
            self.loginStatus.hide()
        else:
            self.loginStatus.show()

        event.accept()

    def closeEvent(self, event: QtGui.QCloseEvent):
        """Save window state and session when app closed.

        Args:
            event (QtGui.QCloseEvent): Close event
        """
        self.settings.size = self.size()
        self.settings.pos = self.pos()
        self.settings.state = self.windowState()
        self.settings.toolbarArea = self.toolBarArea(self.toolBar)

        self.tab.save()

        event.accept()



from PyQt5 import QtWidgets, QtGui

from skippy.utils import translator

import skippy.config

from typing import Callable
import os


def UpdaterDialog(version: str, update_func: Callable[[], None]):
    msgBox = QtWidgets.QMessageBox()
    msgBox.setText(
        translator.Translator()
        .translate("DIALOG.NEW_VERSION_AVAILABLE_LABEL")
        .format(version)
    )
    msgBox.setInformativeText(
        translator.Translator().translate("DIALOG.YOU_CAN_DOWNLOAD_IT_LABEL")
    )

    msgBox.addButton(
        translator.Translator().translate("DIALOG.CANCEL_BUTTON"),
        QtWidgets.QMessageBox.RejectRole,
    )
    installButton = msgBox.addButton(
        translator.Translator().translate("DIALOG.INSTALL_BUTTON"),
        QtWidgets.QMessageBox.ActionRole,
    )

    msgBox.setWindowTitle(f"Skippy - {skippy.config.version}")
    msgBox.setWindowIcon(
        QtGui.QIcon(os.path.join(skippy.config.RESOURCES_FOLDER, "skippy.ico"))
    )

    msgBox.exec_()

    if msgBox.clickedButton() == installButton:
        update_func()
        return False
    return True



######################################################################
### From https://github.com/gmarull/qtmodern
######################################################################

from PyQt5 import QtGui

from skippy.gui import utils

import skippy.config

import os

_STYLESHEET = os.path.join(skippy.config.RESOURCES_FOLDER, "stylesheet", "style.qss")
""" str: Main stylesheet. """


def _apply_base_theme(app):
    """Apply base theme to the application.
    Args:
            app (QApplication): QApplication instance.
    """

    app.setStyle("Fusion")

    with open(_STYLESHEET) as stylesheet:
        app.setStyleSheet(stylesheet.read())


def dark():
    """Apply Dark Theme to the Qt application instance."""
    app = utils.getApplication()

    darkPalette = QtGui.QPalette()

    # base
    darkPalette.setColor(QtGui.QPalette.WindowText, QtGui.QColor(180, 180, 180))
    darkPalette.setColor(QtGui.QPalette.Button, QtGui.QColor(53, 53, 53))
    darkPalette.setColor(QtGui.QPalette.Light, QtGui.QColor(180, 180, 180))
    darkPalette.setColor(QtGui.QPalette.Midlight, QtGui.QColor(90, 90, 90))
    darkPalette.setColor(QtGui.QPalette.Dark, QtGui.QColor(35, 35, 35))
    darkPalette.setColor(QtGui.QPalette.Text, QtGui.QColor(180, 180, 180))
    darkPalette.setColor(QtGui.QPalette.BrightText, QtGui.QColor(180, 180, 180))
    darkPalette.setColor(QtGui.QPalette.ButtonText, QtGui.QColor(180, 180, 180))
    darkPalette.setColor(QtGui.QPalette.Base, QtGui.QColor(42, 42, 42))
    darkPalette.setColor(QtGui.QPalette.Window, QtGui.QColor(53, 53, 53))
    darkPalette.setColor(QtGui.QPalette.Shadow, QtGui.QColor(20, 20, 20))
    darkPalette.setColor(QtGui.QPalette.Highlight, QtGui.QColor(42, 130, 218))
    darkPalette.setColor(QtGui.QPalette.HighlightedText, QtGui.QColor(180, 180, 180))
    darkPalette.setColor(QtGui.QPalette.Link, QtGui.QColor(56, 252, 196))
    darkPalette.setColor(QtGui.QPalette.AlternateBase, QtGui.QColor(66, 66, 66))
    darkPalette.setColor(QtGui.QPalette.ToolTipBase, QtGui.QColor(53, 53, 53))
    darkPalette.setColor(QtGui.QPalette.ToolTipText, QtGui.QColor(180, 180, 180))
    darkPalette.setColor(QtGui.QPalette.LinkVisited, QtGui.QColor(80, 80, 80))

    # disabled
    darkPalette.setColor(
        QtGui.QPalette.Disabled, QtGui.QPalette.WindowText, QtGui.QColor(127, 127, 127)
    )
    darkPalette.setColor(
        QtGui.QPalette.Disabled, QtGui.QPalette.Text, QtGui.QColor(127, 127, 127)
    )
    darkPalette.setColor(
        QtGui.QPalette.Disabled, QtGui.QPalette.ButtonText, QtGui.QColor(127, 127, 127)
    )
    darkPalette.setColor(
        QtGui.QPalette.Disabled, QtGui.QPalette.Highlight, QtGui.QColor(80, 80, 80)
    )
    darkPalette.setColor(
        QtGui.QPalette.Disabled,
        QtGui.QPalette.HighlightedText,
        QtGui.QColor(127, 127, 127),
    )

    app.setPalette(darkPalette)

    _apply_base_theme(app)


def light():
    """Apply Light Theme to the Qt application instance."""
    app = utils.getApplication()

    lightPalette = QtGui.QPalette()

    # base
    lightPalette.setColor(QtGui.QPalette.WindowText, QtGui.QColor(0, 0, 0))
    lightPalette.setColor(QtGui.QPalette.Button, QtGui.QColor(240, 240, 240))
    lightPalette.setColor(QtGui.QPalette.Light, QtGui.QColor(180, 180, 180))
    lightPalette.setColor(QtGui.QPalette.Midlight, QtGui.QColor(200, 200, 200))
    lightPalette.setColor(QtGui.QPalette.Dark, QtGui.QColor(225, 225, 225))
    lightPalette.setColor(QtGui.QPalette.Text, QtGui.QColor(0, 0, 0))
    lightPalette.setColor(QtGui.QPalette.BrightText, QtGui.QColor(0, 0, 0))
    lightPalette.setColor(QtGui.QPalette.ButtonText, QtGui.QColor(0, 0, 0))
    lightPalette.setColor(QtGui.QPalette.Base, QtGui.QColor(237, 237, 237))
    lightPalette.setColor(QtGui.QPalette.Window, QtGui.QColor(240, 240, 240))
    lightPalette.setColor(QtGui.QPalette.Shadow, QtGui.QColor(20, 20, 20))
    lightPalette.setColor(QtGui.QPalette.Highlight, QtGui.QColor(76, 163, 224))
    lightPalette.setColor(QtGui.QPalette.HighlightedText, QtGui.QColor(0, 0, 0))
    lightPalette.setColor(QtGui.QPalette.Link, QtGui.QColor(0, 162, 232))
    lightPalette.setColor(QtGui.QPalette.AlternateBase, QtGui.QColor(225, 225, 225))
    lightPalette.setColor(QtGui.QPalette.ToolTipBase, QtGui.QColor(240, 240, 240))
    lightPalette.setColor(QtGui.QPalette.ToolTipText, QtGui.QColor(0, 0, 0))
    lightPalette.setColor(QtGui.QPalette.LinkVisited, QtGui.QColor(222, 222, 222))

    # disabled
    lightPalette.setColor(
        QtGui.QPalette.Disabled, QtGui.QPalette.WindowText, QtGui.QColor(115, 115, 115)
    )
    lightPalette.setColor(
        QtGui.QPalette.Disabled, QtGui.QPalette.Text, QtGui.QColor(115, 115, 115)
    )
    lightPalette.setColor(
        QtGui.QPalette.Disabled, QtGui.QPalette.ButtonText, QtGui.QColor(115, 115, 115)
    )
    lightPalette.setColor(
        QtGui.QPalette.Disabled, QtGui.QPalette.Highlight, QtGui.QColor(190, 190, 190)
    )
    lightPalette.setColor(
        QtGui.QPalette.Disabled,
        QtGui.QPalette.HighlightedText,
        QtGui.QColor(115, 115, 115),
    )

    app.setPalette(lightPalette)

    _apply_base_theme(app)



"""Configures various variables used by Skippy.

Attributes:
    LANG_FOLDER (str): Language folder
    LOGS_FOLDER (str): Logs folder
    PLUGINS_FOLDER (str): Plugins folder
    PROPERTY_FOLDER (str): Property folder
    RESOURCES_FOLDER (str): Resources folder
    SKIPPY_FOLDER (str): Skippy folder
    version (str): Skippy version same as skippy.__version__
"""
from skippy import __version__

from skippy.utils import is_frozen

import os

version = __version__

SKIPPY_FOLDER = os.path.dirname(os.path.realpath(__file__))

RESOURCES_FOLDER = os.path.join(SKIPPY_FOLDER, "resources")

PROPERTY_FOLDER = os.path.join(SKIPPY_FOLDER, "property")

LOGS_FOLDER = os.path.join(SKIPPY_FOLDER, "logs")

LANG_FOLDER = os.path.join(SKIPPY_FOLDER, "lang")

PLUGINS_FOLDER = os.path.join(SKIPPY_FOLDER, "plugins")

if is_frozen():
    APPDATA_FOLDER = os.path.join(os.getenv("APPDATA"), "Skippy")

    PROPERTY_FOLDER = os.path.join(APPDATA_FOLDER, "property")

    LOGS_FOLDER = os.path.join(APPDATA_FOLDER, "logs")

    PLUGINS_SETTINGS_FOLDER = os.path.join(APPDATA_FOLDER, "plugins")

PLUGINS_SETTINGS_FOLDER = os.path.join(PLUGINS_FOLDER, "settings")



"""Standard directories
"""
from skippy.utils import is_frozen

import skippy.config

import os


def initdirs():
    """Init standard dirs"""
    if is_frozen():
        makedir(skippy.config.APPDATA_FOLDER)
    makedir(skippy.config.LOGS_FOLDER)
    makedir(skippy.config.PROPERTY_FOLDER)
    makedir(skippy.config.PLUGINS_SETTINGS_FOLDER)


def makedir(path: str):
    """Make a directory if it doesn't exist

    Args:
        path (str): Path to folder
    """
    if not os.path.isdir(path):
        os.makedirs(path)



"""Elements core module

Attributes:
    elements (List[AbstractElement]): List of elements
"""
from skippy.api import Field

from abc import ABCMeta, abstractmethod
from typing import Dict, List


class AbstractElement(metaclass=ABCMeta):

    """Abstract element class

    Attributes:
        required_fields (List[Field]): List of element fields
    """

    __alias__: str
    __description__: str

    base: str

    def __init__(self):
        """Initializing element"""
        self.required_fields: List[Field] = []
        self._init_fields()

    def check_args(self, args: Dict[str, str]):
        """Check required arguments

        Args:
            args (Dict[str, str]): Arguments for fields

        Raises:
            TypeError: Don't received required argument
        """
        for field in self.required_fields:
            if field.tag not in args:
                raise TypeError(
                    f"{self.__class__.__name__}.preview() don't received required argument: {field.tag}"
                )

    @abstractmethod
    def _init_fields(self):
        """Abstract init fields method"""
        pass

    def add_field(self, tag: str, name: str = "", description: str = ""):
        """Add field to element

        Args:
            tag (str): Field tag
            name (str, optional): Field name
            description (str, optional): Field description
        """
        self.required_fields.append(Field(tag, name, description))

    def prepare_args(self, args: Dict[str, str]) -> Dict[str, str]:
        """Prepare arguments

        Args:
            args (Dict[str, str]): Arguments for fields

        Returns:
            Dict[str, str]: Prepared arguments
        """
        return args

    def preview(self, args: Dict[str, str]) -> str:
        """Preview element using arguments

        Args:
            args (Dict[str, str]): Arguments for fields

        Returns:
            str: Description
        """
        self.check_args(args)
        args = self.prepare_args(args)

        data = self.base
        for name in args:
            data = data.replace(f"<<{name}>>", args[name])
        return data


class AbstractComponent(AbstractElement):

    """Abstract included component"""

    component: str
    base: str = "[[include <<component>> <<args>>]]"

    multiline: bool = True

    def prepare_args(self, args: Dict[str, str]) -> Dict[str, str]:
        """Prepare arguments

        Args:
            args (Dict[str, str]): Arguments for fields

        Returns:
            Dict[str, str]: Prepared arguments
        """
        next_line = "\n" if self.multiline else " "
        return {
            "component": self.component,
            "args": f" |{next_line}".join(
                [f"{arg}={args[arg]}" for arg in args if args[arg]]
            ),
        }


class BaseImageBlock(AbstractComponent):

    """Base image block"""

    component: str = "component:image-block"

    multiline: bool = False

    align: str

    def _init_fields(self):
        """Initializing fields"""
        self.add_field(
            "name", "ELEMENTS.BASE_IMAGE.NAME.NAME", "ELEMENTS.BASE_IMAGE.NAME.DESC"
        )
        self.add_field(
            "caption",
            "ELEMENTS.BASE_IMAGE.CAPTION.NAME",
            "ELEMENTS.BASE_IMAGE.CAPTION.DESC",
        )
        self.add_field(
            "width", "ELEMENTS.BASE_IMAGE.WIDTH.NAME", "ELEMENTS.BASE_IMAGE.WIDTH.DESC"
        )
        self.add_field(
            "link", "ELEMENTS.BASE_IMAGE.LINK.NAME", "ELEMENTS.BASE_IMAGE.LINK.DESC"
        )

    def prepare_args(self, args: Dict[str, str]) -> Dict[str, str]:
        """Prepare arguments

        Args:
            args (Dict[str, str]): Arguments for fields

        Returns:
            Dict[str, str]: Prepared arguments
        """
        args["align"] = self.align
        return super(BaseImageBlock, self).prepare_args(args)


class RightImageBlock(BaseImageBlock):

    """Right image block"""

    __alias__: str = "ELEMENTS.RIGHT_IMAGE_BLOCK.ALIAS"
    __description__: str = "ELEMENTS.RIGHT_IMAGE_BLOCK.DESC"

    align: str = ""


class LeftImageBlock(BaseImageBlock):

    """Left image block"""

    __alias__: str = "ELEMENTS.LEFT_IMAGE_BLOCK.ALIAS"
    __description__: str = "ELEMENTS.LEFT_IMAGE_BLOCK.DESC"

    align: str = "left"


class CenterImageBlock(BaseImageBlock):

    """Center image block"""

    __alias__: str = "ELEMENTS.CENTER_IMAGE_BLOCK.ALIAS"
    __description__: str = "ELEMENTS.CENTER_IMAGE_BLOCK.DESC"

    align: str = "center"


class ACSBarComponent(AbstractComponent):

    """Anomaly Classification Bar component"""

    __alias__: str = "ELEMENTS.ACS_BAR.ALIAS"
    __description__: str = "ELEMENTS.ACS_BAR.DESC"

    component: str = "component:anomaly-class-bar-source"

    def _init_fields(self):
        self.add_field(
            "item-number", "ELEMENTS.ACS_BAR.ITEM.NAME", "ELEMENTS.ACS_BAR.ITEM.DESC"
        )
        self.add_field(
            "clearance",
            "ELEMENTS.ACS_BAR.CLEARANCE.NAME",
            "ELEMENTS.ACS_BAR.CLEARANCE.DESC",
        )
        self.add_field(
            "container-class",
            "ELEMENTS.ACS_BAR.CONTAINER.NAME",
            "ELEMENTS.ACS_BAR.CONTAINER.DESC",
        )
        self.add_field(
            "disruption-class",
            "ELEMENTS.ACS_BAR.DISRUPTION.NAME",
            "ELEMENTS.ACS_BAR.DISRUPTION.DESC",
        )
        self.add_field(
            "risk-class", "ELEMENTS.ACS_BAR.RISK.NAME", "ELEMENTS.ACS_BAR.RISK.DESC"
        )
        self.add_field(
            "secondary-class",
            "ELEMENTS.ACS_BAR.SEC_CLASS.NAME",
            "ELEMENTS.ACS_BAR.SEC_CLASS.DESC",
        )
        self.add_field(
            "secondary-icon",
            "ELEMENTS.ACS_BAR.SEC_ICON.NAME",
            "ELEMENTS.ACS_BAR.SEC_ICON.DESC",
        )


elements: List[AbstractElement] = [
    RightImageBlock,
    LeftImageBlock,
    CenterImageBlock,
    ACSBarComponent,
]



"""Logger class

Attributes:
    log (logging.Logger): Skippy logger
    LOG_LEVELS (Dict[str, logging.LEVEL]): Log levels same as logging._nameToLevel
"""
import skippy.config

import datetime
import logging
import os


LOG_LEVELS = logging._nameToLevel


def get_file_handler() -> logging.FileHandler:
    """Get file handler

    Returns:
        logging.FileHandler: File handler
    """
    file_handler = logging.FileHandler(
        os.path.join(
            skippy.config.LOGS_FOLDER,
            f"{datetime.datetime.today().strftime('%Y-%m-%d-%H-%M-%S')}.log",
        ),
        "a",
        "utf-8",
        True,
    )
    file_handler.setFormatter(
        logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s", "%Y-%m-%d %H:%M:%S"
        )
    )
    file_handler.setLevel(logging.DEBUG)

    return file_handler


def get_stream_handler() -> logging.StreamHandler:
    """Get stream handler

    Returns:
        logging.StreamHandler: Stream handler
    """
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(
        logging.Formatter(
            "%(asctime)s - %(levelname)s - %(message)s", "%Y-%m-%d %H:%M:%S"
        )
    )

    return stream_handler


def get_logger(name: str) -> logging.Logger:
    """Get logger by name

    Args:
        name (str): Logger name

    Returns:
        logging.Logger: Logger class
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)

    logger.addHandler(get_stream_handler())
    logger.addHandler(get_file_handler())

    return logger


log = get_logger("skippy")



"""Autoupdater classes
"""
from skippy.api import ignore

from skippy.utils import cached_property, is_frozen

import skippy.config

from typing import Tuple, Union, List, Dict, Any
from abc import ABCMeta, abstractmethod
import requests


RequestException = requests.exceptions.RequestException


def version2tuple(version: str) -> Tuple[int]:
    """Convert string version to tuple

    Args:
        version (str): String version

    Returns:
        Tuple[int]: Tuple version
    """
    return tuple(map(int, version.split(".")))


class AbstractUpdateClient(metaclass=ABCMeta):

    """Abstract update client"""

    _apiEndpoint: str

    @staticmethod
    def getClient() -> "AbstractUpdateClient":
        """Get update client for Skippy

        Returns:
            AbstractUpdateClient: Update client
        """
        if is_frozen():
            return GithubReleasesClient()
        return PyPIClient()

    @cached_property
    @ignore(RequestException, {})
    def data(self) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        """Update data property

        Returns:
            Dict[str, Any]: Update data
        """
        response = requests.get(self._apiEndpoint)
        return response.json()

    def checkVersion(self) -> bool:
        """Check if Skippy has a new version

        Returns:
            bool: Is new version available
        """
        return not version2tuple(skippy.config.version) >= version2tuple(self.version)

    @property
    @abstractmethod
    def version(self):
        """Abstract version property"""
        pass

    @abstractmethod
    def update(self):
        """Abstract update method"""
        pass


class PyPIClient(AbstractUpdateClient):

    """PyPI update client"""

    _apiEndpoint = "https://pypi.org/pypi/skippy-pad/json"

    @property
    @ignore(KeyError, "0.0.0")
    def version(self) -> str:
        """Update version property

        Returns:
            str: Version
        """
        return self.data["info"]["version"]

    def update(self):
        """Update method"""
        import os

        os.system("start cmd /c python -m pip install skippy-pad -U")


class GithubReleasesClient(AbstractUpdateClient):

    """Github releases client"""

    _apiEndpoint = "https://api.github.com/repos/skippy-dev/skippy/releases"

    @property
    @ignore(KeyError, {})
    def data(self) -> Dict[str, Any]:
        """Update data property

        Returns:
            Dict[str, Any]: Update data
        """
        return super().data[0]

    @property
    @ignore(KeyError, "0.0.0")
    def version(self) -> str:
        """Update version property

        Returns:
            str: Version
        """
        return self.data["tag_name"][1:]

    def update(self):
        """Update method"""
        import webbrowser

        webbrowser.open(self.data["html_url"])



"""Translator
"""
from skippy.api import Singleton, Language
import skippy.config

from typing import List
import toml
import os


class Translator(metaclass=Singleton):

    """Translator class"""

    def __init__(self):
        """Init translator class"""
        self._language = None

    @staticmethod
    def languages() -> List[str]:
        """Get all exist skippy languages

        Returns:
            List[str]: List of languages code
        """
        return [
            os.path.splitext(file)[0]
            for file in os.listdir(skippy.config.LANG_FOLDER)
            if os.path.isfile(os.path.join(skippy.config.LANG_FOLDER, file))
            if os.path.splitext(file)[1] == ".toml"
        ]

    @staticmethod
    def getLangName(lang: str) -> str:
        """Get lang name by code

        Args:
            lang (str): Language code

        Returns:
            str: Language name
        """
        path = os.path.join(skippy.config.LANG_FOLDER, f"{lang}.toml")
        with open(path, encoding="utf-8") as f:
            return toml.load(f)["LANGUAGE"]

    def load(self, lang: str = "en") -> Language:
        """Load language by code

        Args:
            lang (str, optional): Language code

        Returns:
            Language: Language namedtuple, with language code and dictionary
        """
        path = os.path.join(skippy.config.LANG_FOLDER, f"{lang}.toml")
        with open(path, encoding="utf-8") as f:
            dictionary = toml.load(f)
        self._language = Language(lang, dictionary)

        return self._language

    def translate(self, context: str) -> str:
        """Translate context text

        Args:
            context (str): Text context

        Returns:
            str: Translated string if available, else return input context
        """
        try:
            out = self._language.dictionary
            for text in context.split("."):
                out = out[text]
            return out
        except KeyError:
            return context



from typing import Any
import setuptools
import ast
import re
import os

try:
    BASEDIR = os.path.dirname(os.path.realpath(__file__))
except NameError:
    BASEDIR = None


def read_file(name: str) -> str:
    """Get the string contained in the file named name.

    Args:
        name (str): File name

    Returns:
        str: File content
    """
    with open(name, "r", encoding="utf8") as f:
        return f.read()


def _get_constant(name: str) -> Any:
    """Read a __magic__ constant from skippy/__init__.py.

    We don't import Skippy here because it can go wrong for multiple
    reasons. Instead we use re/ast to get the value directly from the source
    file.

    Args:
        name: The name of the argument to get.

    Return:
        The value of the argument.
    """
    field_re = re.compile(r"__{}__\s+=\s+(.*)".format(re.escape(name)))
    path = os.path.join(BASEDIR, "skippy", "__init__.py")
    line = field_re.search(read_file(path)).group(1)
    value = ast.literal_eval(line)
    return value


readme = read_file("README.md")
requirements = read_file("requirements.txt").splitlines()

setuptools.setup(
    name="skippy-pad",
    version=_get_constant("version"),
    description=_get_constant("description"),
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/skippy-dev/skippy/",
    author=_get_constant("author"),
    author_email=_get_constant("email"),
    python_requires=">=3.7",
    license=_get_constant("license"),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords=["scp", "wikidot", "pyscp", "skippy"],
    packages=setuptools.find_packages(),
    install_requires=requirements,
    package_dir={"skippy": "skippy"},
    package_data={
        "skippy": [
            "resources/*",
            "resources/*/*",
            "lang/*.toml"
        ]
    },
    entry_points={
        "console_scripts": [
            "skippy = skippy.app:run",
        ],
    },
)



from PyQt5 import QtWebEngineWidgets, QtWidgets, QtCore, QtGui

from skippy.api import PageData

from skippy.core import preview

from skippy.gui import thread, utils

import skippy.config

from typing import Optional
import tempfile
import os


class PreviewerWorker(thread.AbstractWorker):
    finished = QtCore.pyqtSignal(str)

    def __init__(self, pdata: PageData):
        super(PreviewerWorker, self).__init__()
        self.pdata = pdata

    def run(self):
        self.finished.emit(preview.render(self.pdata))


class Previewer(QtWidgets.QDialog):
    def __init__(self, pdata: PageData, parent: Optional[QtWidgets.QWidget] = None):
        super(Previewer, self).__init__(parent)
        self._layout = QtWidgets.QVBoxLayout(self)

        self.webEngineView = QtWebEngineWidgets.QWebEngineView(self)

        self._thread = thread.Thread(PreviewerWorker(pdata))
        self._thread.worker.finished.connect(self.load)
        self._thread.start()

        self._layout.addWidget(self.webEngineView)

        self.setLayout(self._layout)

        self.setWindowTitle(f"Skippy - {skippy.config.version}")
        self.setWindowIcon(
            QtGui.QIcon(os.path.join(skippy.config.RESOURCES_FOLDER, "skippy.ico"))
        )

        mainwindow = utils.getMainWindow()

        self.move(mainwindow.x(), mainwindow.y())
        self.resize(mainwindow.width(), mainwindow.height())
        self.setWindowState(mainwindow.windowState())

        self.exec_()

    def load(self, html: str):
        with tempfile.NamedTemporaryFile(
            delete=False, suffix=".html", mode="w", encoding="utf-8"
        ) as tmp:
            tmp.write(html)
            self.webEngineView.load(QtCore.QUrl.fromLocalFile(tmp.name))



from skippy.config import RESOURCES_FOLDER, version
from skippy import __description__

from cx_Freeze import setup, Executable
import platform
import shutil
import sys
import os

path = os.path.dirname(os.path.realpath(__file__))

base = "Win32GUI" if sys.platform == "win32" else None

options = {
    "build_exe": {
        "includes": "atexit",
    }
}

executables = [
    Executable(
        os.path.join(path, "run.py"),
        icon=os.path.join(RESOURCES_FOLDER, "skippy.ico"),
        base=base,
    )
]

setup(
    name="Skippy",
    version=version,
    description=__description__,
    options=options,
    executables=executables,
)

shutil.make_archive(
    os.path.join("output", f"{platform.system()}.standalone"),
    "zip",
    os.path.join("build", "exe.win-amd64-3.9"),
)



import sys

try:
    from functools import cached_property
except ImportError:
    from functools import lru_cache

    cached_property = lambda func: property(lru_cache(func))


def is_frozen() -> bool:
    """Is Skippy compiled to exe

    Returns:
        bool: Compiled or not
    """
    return getattr(sys, "frozen", False)


__all__ = ["cached_property", "is_frozen"]



"""Initialization of Skippy and application-wide things.
"""
from skippy.gui import start_ui

from skippy.core import plugins

from skippy.utils import logger, standarddir, discord_rpc, excepthook
import skippy.config

from prettytable import PrettyTable
import argparse
import sys


def get_argparser() -> argparse.ArgumentParser:
    """Get the argument parser.

    Returns:
        argparse.ArgumentParser: Argument parser
    """
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="print the version - will not run the ui",
    )
    parser.add_argument(
        "-p",
        "--plugins",
        action="store_true",
        help="print plugin list - will not run the ui",
    )
    parser.add_argument(
        "--logging_level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="the level to use for logging - defaults to INFO",
        default="INFO",
    )

    return parser


class App:

    """Application class

    Attributes:
        args (argparse.Namespace): Command-line arguments
    """

    def __init__(self, args: argparse.Namespace):
        """Initializing application class

        Args:
            args (argparse.Namespace): Command-line arguments
        """
        self.args = args

    def run(self):
        """Run application by command-line arguments"""
        if self.args.version:
            self.version()
        elif self.args.plugins:
            self.plugins()
        else:
            self.start_ui()

    def version(self):
        """Print Skippy version"""
        print(f"skippy v{skippy.config.version}")

    def plugins(self):
        """Print plugins list with additional data"""
        table = PrettyTable()
        table.field_names = ["Alias", "Description", "Author", "Version"]

        for plugin in plugins.PluginLoader.plugins_data():
            table.add_row(
                [
                    plugin["__alias__"],
                    plugin["__description__"],
                    plugin["__author__"],
                    plugin["__version__"],
                ]
            )

        print(table)

    def start_ui(self) -> int:
        """Initialize everything and run the application.

        Returns:
            int: Exit code
        """
        logger.log.setLevel(logger.LOG_LEVELS[self.args.logging_level])

        standarddir.initdirs()
        logger.log.info("Initializing directories...")

        logger.log.info("Initializing plugin loader...")
        pluginLoader = plugins.PluginLoader()

        logger.log.info("Load plugins...")
        pluginLoader.load_plugins()

        logger.log.info("Start plugins...")
        pluginLoader.start_plugins()

        logger.log.info("Initializing Discord RPC...")
        rpc = discord_rpc.DiscordRPC()

        logger.log.info("Connect Discord RPC...")
        rpc.connect()

        logger.log.info("Initializing application...")
        exit_code = start_ui()

        logger.log.info("Stop plugins...")
        pluginLoader.stop_plugins()

        logger.log.info("Stop Discord RPC...")
        rpc.close()

        return exit_code


def run():
    """Initialize everything and run the application."""
    excepthook.init()

    parser = get_argparser()

    app = App(parser.parse_args())
    sys.exit(app.run())



from PyQt5 import QtWidgets, QtCore, QtGui

from skippy.api import InlineSyntax, BlockSyntax

from skippy.gui import actionbar, workers, thread, utils

from typing import Optional
import re


inlinePatterns = [
    InlineSyntax("[", "]", True),
    InlineSyntax("//", "//", False),
    InlineSyntax("**", "**", False),
    InlineSyntax("{{", "}}", False),
    InlineSyntax("--", "--", False),
    InlineSyntax("__", "__", False),
    InlineSyntax("^^", "^^", False),
    InlineSyntax(",,", ",,", False),
    InlineSyntax("@@", "@@", False),
    InlineSyntax("@<", ">@", False),
    InlineSyntax("##", "|text##", False),
    InlineSyntax("%%", "%%", False),
    InlineSyntax("{$", "}", False),
    InlineSyntax("%%form_data{", "}", False),
    InlineSyntax("%%form_raw{", "}", False),
    InlineSyntax("%%form_hint{", "}", False),
    InlineSyntax("%%form_label{", "}", False),
    InlineSyntax("%%content{", "}", False),
]

blockPatterns = {
    "module": BlockSyntax(
        True,
        False,
        {
            "Rate": {},
            "ListPages": {
                "pagetype": "normal|hidden|*",
                "category": ".|*",
                "tags": "-|=|==",
                "parent": "-|=|-=|.",
                "link_to": ".",
                "created_at": "=|yyyy|yyyy.mm|last n unit",
                "created_by": "=|-=",
                "rating": "n|=",
                "offset": "n",
                "range": ".|before|after|others",
                "order": "property desc",
                "limit": "n",
                "perPage": "n",
                "reverse": "yes|no",
                "separate": "yes|no",
                "prependLine": "str",
                "appendLine": "str",
                "rss": "str",
                "rssDescription": "str",
                "rssHome": "uri",
                "rssLimit": "n",
                "rssOnly": "true",
                "urlAttrPrefix": "str",
            },
            "ListUsers": {"users": "."},
            "CSS": {"show": "true|false", "disable": "true|false"},
            "Redirect": {"destination": "url"},
            "ListDrafts": {"pageType": "exists|notexists"},
            "CountPages": {
                "pagetype": "normal|hidden|*",
                "category": ".|*",
                "tags": "-|=|==",
                "parent": "-|=|-=|.",
                "link_to": ".",
                "created_at": "=|yyyy|yyyy.mm|last n unit",
                "created_by": "=|-=",
                "rating": "n|=",
                "offset": "n",
                "range": ".|before|after|others",
                "prependLine": "",
                "appendLine": "",
                "urlAttrPrefix": "",
            },
            "TagCloud": {
                "mode": "3d",
                "maxFontSize": "px|pt|em|%",
                "minFontSize": "px|pt|em|%",
                "maxColor": "RRR,GGG,BBB",
                "minColor": "RRR,GGG,BBB",
                "limit": "int",
                "target": "unix_pagename",
                "category": "category",
                "showHidden": "true|false",
                "urlAttrPrefix": "str",
                "skipCategoryFromUrl": "true|false",
                "width": "int",
                "height": "int",
            },
            "PageCalendar": {
                "category": "*|category1,category2",
                "tags": "tag1,tag2",
                "startPage": "unix_pagename",
                "targetPage": "unix_pagename",
                "urlAttrPrefix": "str",
            },
            "PageTree": {
                "root": "unix_pagename",
                "showRoot": "true|false",
                "depth": "int",
            },
            "Backlinks": {},
            "WantedPages": {},
            "OrphanedPages": {},
            "Categories": {},
            "Watchers": {"noActions": "True|false"},
            "Members": {
                "group": "members|admins|moderators",
                "showSince": "no|yes",
                "order": "userID|joined|name[Desc]",
            },
            "Join": {"button": "Join this site!", "class": "my-join-button"},
            "SendInvitations": {},
            "WhoInvited": {},
            "NewPage": {
                "category": "category",
                "template": "unix_pagename",
                "size": "int",
                "button": "create page",
                "format": "/^Reg?Exp$/",
                "tags": "tag1 tag2",
                "parent": "unix_pagename",
                "mode": "edit|save-and-refresh|save-and-go",
                "goTo": "unix_pagename",
            },
            "ThemePreviewer": {"noUi": "true|false"},
            "MailForm": {
                "to": "user1,user2",
                "button": "str",
                "format": "csv",
                "title": "Email Title",
                "successPage": "unix_pagename",
            },
            "PetitionAdmin": {},
            "SiteGrid": {"limit": "int"},
            "FeaturedSite": {},
            "Feed": {
                "src": "http://path.to/feed.xml",
                "limit": "int",
                "offset": "int",
            },
            "FrontForum": {
                "category": "12,102",
                "feed": "str",
                "feedTitle": "sitename feed",
                "limit": "int",
                "offset": "int",
                "fixRelativeLinks": "true|false",
            },
            "Comments": {
                "title": "Comments",
                "hide": "true|false",
                "hideForm": "true|false",
                "order": "reverse|forwards",
            },
            "RecentPosts": {},
            "MiniRecentThreads": {"limit": "int"},
            "MiniActiveThreads": {"limit": "int"},
            "MiniRecentPosts": {"limit": "int"},
            "RatedPages": {
                "category": "category-name",
                "order": "date-created-|rating-[desc|asc]",
                "minRating": "int",
                "maxRating": "int",
                "limit": "int",
                "comments": "True|false",
            },
            "FlickrGallery": {
                "userName": "flickr-user",
                "tags": "tag1, tag2",
                "tagMode": "all|any",
                "sort": "date-posted-|date-taken-|interestingness-[desc|asc]|relevance",
                "photosetId": "photoset-id",
                "groupId": "group-name",
                "groupUrl": "flickr-url",
                "perPage": "int (1-100)",
                "limitPages": "int",
                "size": "square|thumbnail|small|medium",
                "disableBrowsing": "true|false",
                "contentType": "photos|screenshots|other|all",
            },
            "Files": {},
            "Search": {"mini": "true", "a": "p|pf|f"},
            "SearchAll": {},
            "SearchUsers": {},
            "SiteChanges": {},
            "ManageSite": {},
            "Clone": {"source": ".|site-unix-name", "button": "str"},
        },
    ),
    "collapsible": BlockSyntax(
        True,
        False,
        {
            "": {
                "show": "str",
                "hide": "str",
                "folded": "no|yes",
                "hideLocation": "both|bottom|top",
            }
        },
    ),
    "code": BlockSyntax(
        True,
        False,
        {"": {"type": "lang"}},
    ),
    "iftags": BlockSyntax(True, False),
    "toc": BlockSyntax(False, False),
    "include": BlockSyntax(False, False),
    "user": BlockSyntax(False, False),
    "*user": BlockSyntax(False, False),
    "html": BlockSyntax(True, False),
    "gallery": BlockSyntax(
        True,
        False,
        {
            "": {
                "size": "square|thumbnail|small|medium",
                "order": "name|created_at [desc]",
                "viewer": "true|false",
            }
        },
    ),
    "div": BlockSyntax(
        True,
        False,
        {"": {"style": "css: code;", "class": "css-class"}},
    ),
    "span": BlockSyntax(
        True,
        True,
        {"": {"style": "css: code;", "class": "css-class"}},
    ),
    "footnote": BlockSyntax(True, True),
    "size": BlockSyntax(True, True),
    "table": BlockSyntax(
        True,
        False,
        {"": {"style": "css: code;", "class": "css-class"}},
    ),
    "row": BlockSyntax(
        True,
        False,
        {"": {"style": "css: code;", "class": "css-class"}},
    ),
    "cell": BlockSyntax(
        True,
        False,
        {"": {"style": "css: code;", "class": "css-class"}},
    ),
    "image": BlockSyntax(
        False,
        True,
        {
            "": {
                "link": "uri",
                "alt": "str",
                "title": "str",
                "width": "px|em|%",
                "height": "px|em|%",
                "style": "css",
                "class": "css-class",
                "size": "square|thumbnail|small|medium|large",
            },
        },
    ),
    "=": BlockSyntax(True, False),
    ">": BlockSyntax(True, False),
    "<": BlockSyntax(True, False),
    "==": BlockSyntax(True, False),
}


def splitByPosition(line: str, pos: int) -> tuple[str, str]:
    return line[:pos], line[pos:]


class Completer(QtWidgets.QCompleter):
    insertText = QtCore.pyqtSignal(str)

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super(Completer, self).__init__(parent)
        self.setCompletionMode(self.PopupCompletion)
        self.highlighted.connect(self.setSelected)

        self.setWidget(parent)
        self.setCaseSensitivity(QtCore.Qt.CaseInsensitive)

        self.popup().clicked.connect(lambda index: self.insertText.emit(index.data()))
        self.insertText.connect(parent.insertCompletion)

        self.popup().setStyleSheet("font-family:Arial; font-size:10pt;")

    def setSelected(self, text: str):
        self.lastSelected = text

    @property
    def selected(self) -> str:
        return self.lastSelected


class AdvancedEditor(QtWidgets.QPlainTextEdit):
    RE_BLOCK = r"\[\[([\w=><]+)(?:\s(\w+)|\s|)(?:\s(\S+)|)+\]\]"

    fileDragAndDroped = QtCore.pyqtSignal(str, bytes)

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super(AdvancedEditor, self).__init__(parent)
        self.completer = Completer(self)

        self.setAcceptDrops(True)

    def insertCompletion(self, completion: str):
        extra = len(completion) - len(self.completer.completionPrefix())
        if extra:
            self.insertPlainText(completion[-extra:])

    def getLineUnderCursor(self):
        cursor = self.textCursor()
        cursor.select(cursor.BlockUnderCursor)

        return cursor.selectedText().replace("\u2029", "")

    @property
    def acEnabled(self) -> bool:
        return {"true": True, "false": False}[utils.getMainWindow().settings.acEnabled]

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        if (
            self.completer
            and event.key() in utils.ENTER_KEYS
            and self.completer.popup().isVisible()
        ):
            self.completer.insertText.emit(self.completer.selected)
            self.completer.popup().hide()
            return

        curentLine = self.getLineUnderCursor()
        pos = self.textCursor().positionInBlock()

        super().keyPressEvent(event)
        self.suggestingBlockParams()
        self.closeBlocks(curentLine, pos, event)
        self.appendLists(curentLine, event)
        if self.acEnabled:
            self.completeInlineBrackets(event)

    def openCompleterPopup(self, model: list, prefix: str):
        self.completer.setModel(QtCore.QStringListModel(model, self.completer))
        self.completer.setCompletionPrefix(prefix)

        popup = self.completer.popup()
        popup.setCurrentIndex(self.completer.completionModel().index(0, 0))

        rect = self.cursorRect()
        rect.setWidth(
            popup.sizeHintForColumn(0) + popup.verticalScrollBar().sizeHint().width()
        )
        self.completer.complete(rect)

    def suggestingBlockParams(self):
        selected = self.getLineUnderCursor()

        block = re.match(self.RE_BLOCK, selected)
        if not block:
            return

        word = splitByPosition(selected, self.textCursor().positionInBlock())[0]

        blockNamePrefix = block.group(1)

        if word[-len(blockNamePrefix) :] == blockNamePrefix:
            self.openCompleterPopup(list(blockPatterns), blockNamePrefix)
            return

        blockParamNamePrefix = block.group(2)
        if (
            blockNamePrefix in blockPatterns
            and "" in blockPatterns[blockNamePrefix].dataVar
        ):
            params = blockPatterns[blockNamePrefix].dataVar[""]
            if (
                blockParamNamePrefix
                and word[-len(blockParamNamePrefix) :] == blockParamNamePrefix
                and blockNamePrefix in blockPatterns
            ):
                self.openCompleterPopup(
                    [f'{param}="{params[param]}"' for param in params],
                    blockParamNamePrefix,
                )
                return

            blockParamsPrefix = block.group(3)
            if (
                blockParamsPrefix
                and word[-len(blockParamsPrefix) :] == blockParamsPrefix
                and blockNamePrefix in blockPatterns
            ):
                self.openCompleterPopup(
                    [f'{param}="{params[param]}"' for param in params],
                    blockParamsPrefix,
                )
                return
        else:
            if (
                blockParamNamePrefix
                and word[-len(blockParamNamePrefix) :] == blockParamNamePrefix
                and blockNamePrefix in blockPatterns
            ):
                self.openCompleterPopup(
                    list(blockPatterns[blockNamePrefix].dataVar), blockParamNamePrefix
                )
                return

            blockParamsPrefix = block.group(3)
            if (
                blockParamsPrefix
                and word[-len(blockParamsPrefix) :] == blockParamsPrefix
                and blockNamePrefix in blockPatterns
                and blockParamNamePrefix in blockPatterns[blockNamePrefix].dataVar
            ):
                params = blockPatterns[blockNamePrefix].dataVar[blockParamNamePrefix]
                self.openCompleterPopup(
                    [f'{param}="{params[param]}"' for param in params],
                    blockParamsPrefix,
                )
                return

    def closeBlocks(self, currentLine: str, pos: int, event: QtGui.QKeyEvent):
        left = splitByPosition(currentLine, pos)[0]

        block = re.match(self.RE_BLOCK, left)
        if block:
            for key in blockPatterns:
                pattern = blockPatterns[key]
                if block.group(1) == key and pattern.closeBlock:
                    closeBlock = f"[[/{key}]]"
                    if event.key() in utils.ENTER_KEYS and not pattern.inline:
                        self.insertPlainText("\n" + closeBlock)

                        cursor = self.textCursor()
                        cursor.movePosition(cursor.Up)
                        self.setTextCursor(cursor)
                    elif (
                        event.text()
                        and event.key() not in utils.ENTER_KEYS
                        and event.key() not in utils.DELETE_KEYS
                        and pattern.inline
                        and closeBlock not in currentLine
                    ):
                        self.insertPlainText(closeBlock)

                        cursor = self.textCursor()
                        cursor.movePosition(cursor.Left, n=len(closeBlock))
                        self.setTextCursor(cursor)

    def appendLists(self, currentLine: str, event: QtGui.QKeyEvent):
        if (
            currentLine[:2] in ("* ", "# ")
            and currentLine[:2] != currentLine
            and event.key() in utils.ENTER_KEYS
        ):
            self.insertPlainText(currentLine[:2])

    def completeInlineBrackets(self, event: QtGui.QKeyEvent):
        selected = self.getLineUnderCursor()

        left, right = splitByPosition(selected, self.textCursor().positionInBlock())
        for pattern in inlinePatterns:
            if (
                event.text() == pattern.opening[-1:]
                and left[-len(pattern.opening) :] == pattern.opening
                and (
                    right[: len(pattern.closing)] != pattern.closing or pattern.multiple
                )
            ):
                self.insertPlainText(pattern.closing)
                cursor = self.textCursor()
                cursor.movePosition(cursor.Left, n=len(pattern.closing))
                self.setTextCursor(cursor)

    def contextMenuEvent(self, event: QtGui.QContextMenuEvent):
        actionbar.ContextMenu(self).exec_(self.mapToGlobal(event.pos()))

    def dragEnterEvent(self, event: QtGui.QDragEnterEvent):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event: QtGui.QDropEvent):
        self._thread = thread.Thread(
            workers.FileWorker([url.toLocalFile() for url in event.mimeData().urls()])
        )
        self._thread.worker.progress.connect(self.fileDragAndDroped.emit)
        self._thread.start()

        dummyEvent = QtGui.QDropEvent(
            event.posF(),
            event.possibleActions(),
            QtCore.QMimeData(),
            event.mouseButtons(),
            event.keyboardModifiers(),
        )
        super(AdvancedEditor, self).dropEvent(dummyEvent)
